#pragma once
#include "../../../Utils/FSM/FSM.h"
#include "../../../Utils/Functors.hpp"
// Some centralized header file inclusions...
#include "../../../Common.h"
#include "../../../PerceptualState/PerceptualState.h"
#include "../../../PerceptualState/Locations/Common.h"
#include "Actions/AttackMoveAction.h"
#include "Actions/BuildAction.h"
#include "Actions/AttackUnitAction.h"
#include "Actions/MoveAction.h"
#include "Actions/StopAction.h"
#include "Actions/LoadUnit.h"
#include "Actions/UnloadAllUnitsAtPosition.h"

class Layer1FSM;

/**
 * @brief A behavior controls the execution of some aspect of the system. It corresponds to a finite 
 *        state machine producing actions and switching states according to the transitions activated
 *        by external events.
 *
 * In order to allow plan recognition, behaviors must conform to some specification (akin to a kind 
 * of reflection schema), so ad hoc code might be problematic. Additionally, make sure all behaviors
 * are instantiated from the heap memory (using new)!
 */

class IBehavior : public virtual IFSM
// protected inheritance to avoid exposing FSM-named methods, will change to public if it causes problems
{
public:
  template <class FSM, class DerivedState> friend class FSMState;
  template <class Behavior, class DerivedState> friend class BehaviorState;
  //typedef boost::transform_iterator<convert_fsm_to_behavior, sub_fsm_const_iterator> sub_behavior_const_iterator;
  //typedef boost::transform_iterator<convert_fsm_to_behavior, sub_fsm_iterator> sub_behavior_iterator;

  /**
   * @brief Constructs a new behavior.
   * @param parent The parent of the new behavior.
   * @param initialPriority The initial priority of the new behavior.
   */
  IBehavior(const Process* parent, unsigned int initialPriority)
    : IFSM(parent)
    , Process(parent)
    , m_id(s_nextId++)
    , m_currentPriority(initialPriority)
  {
  }

  /**
   * @brief Behavior's destructor. All sub-behaviors should actually have been removed before destruction.
   */
  virtual ~IBehavior() {}

  /**
   * @return The name of this behavior.
   */
  virtual std::string getName() const = 0;

  // DON'T CALL THESE METHODS IF A SUB-BEHAVIOR MAY BE ADDED/REMOVED!
  //sub_behavior_const_iterator subBehaviorBegin() const { return sub_behavior_const_iterator(FSM::subFSMBegin()); }
  //sub_behavior_const_iterator subBehaviorEnd() const { return sub_behavior_const_iterator(FSM::subFSMEnd()); }

  /**
   * @return The current priority of this behavior.
   */
  unsigned int getCurrentPriority() const
  {
    return m_currentPriority;
  }

protected:
  virtual void output(std::ostream& out) const
  {
    out << "#" << m_id << " (P" << m_currentPriority << ")- " << getName();
  }

  /**
   * @brief Starts the behavior, which includes executing the code associated with the initial state.
   */
  virtual void executeImpl()
  {
    SPAR_LOG(LogTypes::LAYER_1_DECISION, "Starting behavior: \"%s\"", getName().c_str());
    IFSM::executeImpl();
  }

  virtual void terminateImpl()
  {
    SPAR_LOG(LogTypes::LAYER_1_DECISION, "Terminating behavior: \"%s\"", getName().c_str());
    IFSM::terminateImpl();
  }

  // DON'T CALL THESE METHODS IF A SUB-BEHAVIOR MAY BE ADDED/REMOVED!
  //sub_behavior_iterator subBehaviorBegin() { return sub_behavior_iterator(FSM::subFSMBegin()); }
  //sub_behavior_iterator subBehaviorEnd() { return sub_behavior_iterator(FSM::subFSMEnd()); }

  template <class Function>
  Function for_each_sub_behavior(Function function)
  {
    return for_each_child_process(function);
  }

  //
  // Variables
  //

  /**
   * @brief The current priority of this behavior.
   */
  unsigned int m_currentPriority;

  const unsigned int m_id;

private:
  static unsigned int s_nextId;

  IBehavior(const IBehavior& other);
};

template <class DerivedBehavior>
class Behavior : public virtual IBehavior, public FSM<DerivedBehavior>
{
public:
  Behavior(const Process* parent, unsigned int initialPriority)
    : IBehavior(parent, initialPriority)
    , FSM(parent)
    , IFSM(parent)
    , Process(parent)
  {
  }
};