#pragma once
#include "../../../Common.h"
#include "../../../PersistentUnitGroup.h"
#include "IBehaviorState.h"
#include <BWAPI.h>

class ArbitratorUnitData;

/**
 * @brief Base class for low-level (BWAPI-level) actions.
 */
class LowLevelAction
{
public:
  /**
   * @brief Constructs a new low-level action associated with an existing behavior.
   * @param state the behavior state from which this action originates.
   */
  LowLevelAction(IBehaviorState& state)
    : m_state(state)
  {
  }

  virtual ~LowLevelAction() {}

  /**
   * @return a description of this action.
   */
  virtual std::string getDescription() const = 0;

  /**
   * @brief Executes this action with a particular unit.
   * @param unit the unit which will execute the action.
   * @return whether the unit successfully accomplished the action.
   */
  virtual bool execute(BWAPI::Unit* unit) const = 0;

  /**
   * @brief Executes this action with a particular group of units.
   * @param group the group of units which will execute the action.
   * @return whether all units successfully accomplished the action.
   */
  bool execute(const PersistentUnitGroup* group) const
  {
    bool result = true;
    for (PersistentUnitGroup::const_iterator it = group->begin();
         it != group->end();
         ++it)
    {
      result = execute(*it) && result;
    }
    return result;
  }

  //
  // At the moment an action is considered terminated when the issuing behavior changes state.
  // This could be implemented differently, i.e. each action keeps a list of subscribers
  // and each behavior state is responsible for terminating its actions.
  //

  typedef IBehaviorState::OnStateExit OnActionEnded;
  OnActionEnded& getActionEndedEvent() const { return m_state.getStateExitEvent(); }

protected:
  /**
   * @brief the state from which this action originates.
   */
  IBehaviorState& m_state;
};
