#pragma once
#include "../../../../PerceptualState/Locations/Dynamic/CircularUnitDynamicLocation.h"
#include "../Conditions/IsIdle.h"
#include "../Conditions/AtLeastOneUnit.h"
#include "../UnitBehavior.h"
#include "../BehaviorState.h"
#include "../Arbitrator.h"

/**
 * @brief The AvoidEnemies behavior commands a particular unit to flee when enemies are around.
 */
class AvoidEnemies : public UnitBehavior<AvoidEnemies>
{
protected:
  class Idle;
  class Avoiding;
  class ReadyToAvoid;
  friend class Idle;
  friend class Avoiding;
  friend class ReadyToAvoid;

  /**
   * @brief The Idle state is active when no enemies are around.
   */
  class Idle : public BehaviorState<AvoidEnemies, Idle>
  {
  public:
    Idle(AvoidEnemies& behavior);
    void onEnterImpl();
    void onExitImpl() {}

    void onTimer()
    {
      m_transitionToReadyState.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(Idle, onTimer) TimerEvent;

    typedef boost::tuple<
        TimerEvent&
      > Events;
		const Events m_events;
    
    typedef boost::tuple<
        Transition<Idle, IsUnitIdle, NoTransitionAction, ReadyToAvoid>&
      > Transitions;
		const Transitions m_transitions;

  protected:
    // Variables

    // Actions
    layer1::StopAction m_stopAction;

    // Transitions
    TimerEvent m_timerEvent;
    Transition<Idle, IsUnitIdle, NoTransitionAction, ReadyToAvoid> m_transitionToReadyState;
  };

  /**
   * @brief The ReadyToAvoid state is waiting for enemy to be detected.
   */
  class ReadyToAvoid : public BehaviorState<AvoidEnemies, ReadyToAvoid>
  {
  public:
    ReadyToAvoid(AvoidEnemies& behavior);
    void onEnterImpl();
    void onExitImpl() {}

    void onTimer()
    {
      m_context.m_enemyUnits = Spar->getPerceptualState().getEnemyUnits(CircularUnitDynamicLocation(m_context.m_unit));
      m_transitionToAvoidingState.testTransition(boost::tuple<const UnitGroup&>(m_context.m_enemyUnits));
    }

    typedef SCHEDULER_EVENT(ReadyToAvoid, onTimer) TimerEvent;

    typedef boost::tuple<
        TimerEvent&
      > Events;
		const Events m_events;
    
    typedef boost::tuple<
        Transition<ReadyToAvoid, AtLeastOneUnit, NoTransitionAction, Avoiding, EventHandler1<const UnitGroup&>>&
      > Transitions;
		const Transitions m_transitions;

  protected:
    // Variables

    // Actions

    // Transitions
    TimerEvent m_timerEvent;
    Transition<ReadyToAvoid, AtLeastOneUnit, NoTransitionAction, Avoiding, EventHandler1<const UnitGroup&>> m_transitionToAvoidingState;
  };

  /**
   * @brief The Avoiding state is active when fleeing from enemies.
   */
  class Avoiding : public BehaviorState<AvoidEnemies, Avoiding>
  {
  public:
    Avoiding(AvoidEnemies& behavior);
    void onEnterImpl();
    void onExitImpl() {}

    void onTimer();

    typedef SCHEDULER_EVENT(Avoiding, onTimer) TimerEvent;

    typedef boost::tuple<
        TimerEvent&
      > Events;
		const Events m_events;
    
    typedef boost::tuple<
        Transition<Avoiding, AtLeastOneUnit, NoTransitionAction, Avoiding, EventHandler1<const UnitGroup&>>&,
        Transition<Avoiding, Not<AtLeastOneUnit>, NoTransitionAction, Idle, EventHandler1<const UnitGroup&>>&
      > Transitions;
		const Transitions m_transitions;

  protected:
    // Variables

    // Actions
    layer1::MoveAction m_moveAction;

    // Transitions
    TimerEvent m_timerEvent;
    Transition<Avoiding, AtLeastOneUnit, NoTransitionAction, Avoiding, EventHandler1<const UnitGroup&>> m_transitionToSelf;
    Transition<Avoiding, Not<AtLeastOneUnit>, NoTransitionAction, Idle, EventHandler1<const UnitGroup&>> m_transitionToIdle;
  };

public:
  AvoidEnemies(const Process* parent, unsigned int initialPriority, BWAPI::Unit* avoidingUnit)
    : UnitBehavior(parent, initialPriority, avoidingUnit)
    , IUnitBehavior(parent, initialPriority, avoidingUnit)
    , IBehavior(parent, initialPriority)
    , IFSM(parent)
    , Process(parent)
    , m_idleState(*this), m_avoidingState(*this),m_readyToAvoidState(*this)
    , m_states(m_idleState, m_avoidingState, m_readyToAvoidState)
  {
  }

  virtual std::string getName() const
  {
    return "AvoidEnemies (" + getObjectDescription(m_unit) + ")";
  }

  typedef boost::tuple<
      Idle&,
      Avoiding&,
      ReadyToAvoid&
    > States;
	const States m_states;

protected:
  virtual void executeImpl()
  {
    UnitBehavior::executeImpl();
    m_idleState.onEnter();
  }

  // Arguments

  // Variables
  UnitGroup m_enemyUnits;

  // States
  Idle m_idleState;
  Avoiding m_avoidingState;
  ReadyToAvoid m_readyToAvoidState;
};
