#include "stdafx.h"
#include "Layer3Script.h"
#include <ctime>

using namespace layer2;

Layer3Script* Layer3Script::s_instance;

Layer3Script::Layer3Script(Layer3TacticalRecognition<Layer2Cueing<Layer1GameState>>& layer3TacticalRecognition, Layer2Managers& layer2Managers)
  : Layer3Tactical(layer3TacticalRecognition, layer2Managers)
  , Layer3ScriptTypes(*this)
  , m_unitReleasedEventHandler(*this)
  , m_manageResources(*this)
  , m_buildPylons(*this)
  , m_buildGateways(*this)
  , m_buildExpansions(*this)
  , m_buildCannons(*this)
  , m_scout(*this)
  , m_progressTech(*this)
  , m_recruitArmy(*this)
  , m_manageArmy(*this)
  , m_mainBase(NULL)
{
  s_instance = this;
}

Layer3Script::~Layer3Script()
{
}

void Layer3Script::onStart()
{
  Layer3Tactical::onStart();

  srand(static_cast<unsigned int>(time(NULL)));

  assert(Spar->getPerceptualState().getMap()->getOwnStartBaseLocation() != NULL);
  assert(Spar->getPerceptualState().getMap()->getEnemyPossibleStartBaseLocations().size() > 0);

  m_layer2Actions.getUnitIdleEvent().subscribe(&m_unitReleasedEventHandler, NULL);

  // Set up mining probes at main base
  m_mainBase = dynamic_cast<L3BaseLocation*>(Spar->getPerceptualState().getMap()->getOwnStartBaseLocation());
  assert(m_mainBase != NULL);
  const std::set<BWAPI::Unit*>& units=BWAPI::Broodwar->self()->getUnits();
  for(std::set<BWAPI::Unit*>::const_iterator i = units.begin(); i != units.end(); i++) {
    if((*i)->getType() == BWAPI::UnitTypes::Protoss_Probe) {
      m_mainBase->getCollectMinerals()->addUnit(*i);
    }
  }

  // Activate periodic tasks
  Spar->getScheduler().activate(&m_buildCannons, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 7, 0);
  Spar->getScheduler().activate(&m_buildExpansions, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 10, 1);
  Spar->getScheduler().activate(&m_buildGateways, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 5, 0);
  Spar->getScheduler().activate(&m_buildPylons, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 1, 0);
  Spar->getScheduler().activate(&m_manageArmy, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 10, 4);
  Spar->getScheduler().activate(&m_manageResources, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 12, 0);
  Spar->getScheduler().activate(&m_progressTech, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 10, 2);
  Spar->getScheduler().activate(&m_recruitArmy, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 10, 3);
  Spar->getScheduler().activate(&m_scout, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 20, 0);
}

void Layer3Script::onEnd()
{
  Layer3Tactical::onEnd();
}

void Layer3Script::onUnitRecruited(BWAPI::Unit* unit, StaticLocation* location)
{
  assert(unit->getType() != BWAPI::UnitTypes::Protoss_Probe);

  m_manageArmy.addUnit(unit);
}

void Layer3Script::onUnitReleased(void*, BWAPI::Unit* unit, const layer2::GroupTacticalAction* action)
{
  // This callback takes place when a unit has been released from an action (typically when the action terminates)

  // Could also keep a stack of previous actions, per unit? (for example if a scouting probe was reserved for a building action)

  assert(!unit->getType().isBuilding());

  if (unit->getType() == BWAPI::UnitTypes::Protoss_Probe)
  {
    L3BaseLocation* baseLocation = getNearestOwnMiningBaseLocation(unit->getPosition());
    if (baseLocation != NULL)
      baseLocation->getCollectMinerals()->addUnit(unit);
  }
  else
  {
    m_manageArmy.addUnit(unit);
  }
}

void Layer3Script::onFrame()
{
  Layer3Tactical::onFrame();
}

int Layer3Script::countSupplyForOneRoundTroops() const
{
  const Map* map = Spar->getPerceptualState().getMap();
  int supply = 0;
  supply += map->getNbOwnUnits(BWAPI::UnitTypes::Protoss_Nexus) * BWAPI::UnitTypes::Protoss_Probe.supplyRequired();
  supply += map->getNbOwnUnits(BWAPI::UnitTypes::Protoss_Gateway) * BWAPI::UnitTypes::Protoss_Zealot.supplyRequired();
  supply += map->getNbOwnUnits(BWAPI::UnitTypes::Protoss_Robotics_Facility) * BWAPI::UnitTypes::Protoss_Reaver.supplyRequired();
  supply += map->getNbOwnUnits(BWAPI::UnitTypes::Protoss_Stargate) * BWAPI::UnitTypes::Protoss_Corsair.supplyRequired();
  return supply;
}

Layer3Script::ResourcesSaturation Layer3Script::countProbesSaturationRatio() const
{
  size_t nbTotalCrystals = 0;
  size_t nbTotalGeysers = 0;
  size_t nbTotalProbesMinerals = 0;
  size_t nbTotalProbesGas = 0;
  Map* map = Spar->getPerceptualState().getMap();
  for (Map::base_location_const_iterator it = map->baseLocationsBegin(); it != map->baseLocationsEnd(); ++it)
  {
    const L3BaseLocation* baseLocation = dynamic_cast<L3BaseLocation*>(*it);
    if (baseLocation->getNbCompletedOwnUnits(BWAPI::UnitTypes::Protoss_Nexus) > 0)
    {
      nbTotalCrystals += baseLocation->getLastKnownNbCrystalsLeft();
      nbTotalProbesMinerals += baseLocation->getNbProbesOnMinerals();
      nbTotalGeysers += baseLocation->getNbGeysers();
      nbTotalProbesGas += baseLocation->getNbProbesOnGas();
    }
  }

  double totalRatio = 0.0;
  if (nbTotalCrystals > 0)
  {
    for (Map::base_location_const_iterator it = map->baseLocationsBegin(); it != map->baseLocationsEnd(); ++it)
    {
      const L3BaseLocation* baseLocation = dynamic_cast<L3BaseLocation*>(*it);
      if (baseLocation->getNbOwnUnits(BWAPI::UnitTypes::Protoss_Nexus) > 0)
      {
        const double saturation = baseLocation->probesMineralsSaturation();
        if (saturation != 0.)
        {
          const double ratio = baseLocation->getNbProbesOnMinerals() / saturation;
          totalRatio += (baseLocation->getLastKnownNbCrystalsLeft() * ratio) / double(nbTotalCrystals);
        }
      }
    }
  }
  return ResourcesSaturation(nbTotalProbesMinerals, nbTotalCrystals, totalRatio,
                    nbTotalProbesGas, nbTotalGeysers);
}

L3BaseLocation* Layer3Script::getNearestOwnMiningBaseLocation(BWAPI::Position position) const
{
  // TODO: nearest (walkable) base locations first

  BaseLocation* bestLocation = NULL;
  double minDistance = boost::numeric::bounds<double>::highest();

  const Map* map = Spar->getPerceptualState().getMap();
  for (Map::base_location_const_iterator it = map->baseLocationsBegin(); it != map->baseLocationsEnd(); ++it)
  {
    BaseLocation* candidate = *it;
    if (candidate->getNbCompletedOwnUnits(BWAPI::UnitTypes::Protoss_Nexus) > 0 &&
        candidate->getLastKnownNbCrystalsLeft() > 0)
    {
      double distance = position.getDistance(candidate->getCurrentCenter());
      if (distance < minDistance)
      {
        minDistance = distance;
        bestLocation = candidate;
      }
    }
  }
  return dynamic_cast<L3BaseLocation*>(bestLocation);
}

Layer3Script::ResourcesAcquisitionRatio Layer3Script::computeResourcesAcquisitionRatio() const
{
  // Should depend on current actions (resources requests?) / current units composition
  //m_layer2Actions.getCurrentResourcesRequests()
  //m_recruitArmy.getCurrentUnitsComposition()

  // At some point anyway it is better to invest probes in collecting gas (even if said gas piles up)
  // than to over-saturate a mineral line. The gas can be later dumped on mass high templars / archons.

  // TODO: compute correct ratio

  ResourcesAcquisitionRatio currentRatio = m_manageResources.getTargetResourcesRatio();

  if (currentRatio == ResourcesAcquisitionRatio(2.0, 0.25) || 
      Spar->getPerceptualState().getMap()->getNbOwnUnits(BWAPI::UnitTypes::Protoss_Cybernetics_Core) > 0)
  {
    return ResourcesAcquisitionRatio(2.0, 0.25);
  }
  else
  {
    return ResourcesAcquisitionRatio(2.0, 0.0);
  }
}

void Layer3Script::onDependentTechsUpdate()
{
  // Called from m_progressTech

  updateResourcesRatio();
}
