#include "stdafx.h"
#include "ManageResourcesTask.h"
#include "../Layer3Script.h"

const bool MOVE_PROBES_AWAY_IF_UNDER_ATTACK = true;
const size_t MAX_TOTAL_PROBES = 70;

using layer2::GroupTacticalAction;

ManageResourcesTask::ProbesResourcesDistribution ManageResourcesTask::getTargetProbesResourcesDistribution(const ResourcesSaturation& saturation) const
{
  // TODO: improve calculation (mathematical modelling of resources acquisition rate)

  const size_t totalNbProbes = saturation.getTotalNbProbes();

  const size_t totalNbProbesMinerals = size_t(totalNbProbes * m_targetResourcesRatio.getPctMinerals());
  const size_t totalNbProbesGas = totalNbProbes - totalNbProbesMinerals;

  return ProbesResourcesDistribution(totalNbProbesMinerals, totalNbProbesGas);
}

void ManageResourcesTask::run()
{
  // TODO: island locations... or situations where ground transfer is not feasible
  // TODO: depleted geysers
  // TODO: destroyed nexus or mineral line under attack...

  const Map* map = Spar->getPerceptualState().getMap();
  ResourcesSaturation saturation = Layer3Script::getInstance().countProbesSaturationRatio();
  const double probesPerCrystalRatio = double(saturation.getNbProbesOnMinerals()) / double(saturation.getNbCrystals());

  size_t totalProbesTransferGas = getTargetProbesResourcesDistribution(saturation).getNbProbesOnGas() - saturation.getNbProbesOnGas();

  const bool stopRecruitingProbes = (saturation.getMineralsSaturationRatio() >= 1.0 && saturation.getGasSaturationRatio() >= 1.0)
                                 || saturation.getTotalNbProbes() >= MAX_TOTAL_PROBES;

  struct BaseManagement
  {
    BaseManagement(L3BaseLocation* p_baseLocation, int p_probesDiff, bool p_isUnderAttack)
      : baseLocation(p_baseLocation)
      , probesDiff(p_probesDiff)
      , isUnderAttack(p_isUnderAttack)
    {}
    L3BaseLocation* baseLocation;
    int probesDiff;
    bool isUnderAttack;
  };
  std::list<BaseManagement> basesSaturation;

  //
  // "Sort" mining bases by probes saturation
  //
  for (Map::base_location_const_iterator it = map->baseLocationsBegin(); it != map->baseLocationsEnd(); ++it)
  {
    L3BaseLocation* baseLocation = dynamic_cast<L3BaseLocation*>(*it);
    if (baseLocation->getNbCompletedOwnUnits(BWAPI::UnitTypes::Protoss_Nexus) > 0)
    {
      baseLocation->executeCollectResources();

      if (stopRecruitingProbes)
        baseLocation->cancelRecruitProbes();

      bool isUnderAttack = baseLocation->hasEnemyArmyUnits();
      if (isUnderAttack)
      {
        if (MOVE_PROBES_AWAY_IF_UNDER_ATTACK)
        {
          baseLocation->cancelRecruitProbes();

          // Ghetto fix: transfer gas probes to minerals to send them elsewhere
          GroupTacticalAction::transfer(baseLocation->getCollectGas(), baseLocation->getCollectMinerals(), baseLocation->getCollectGas()->getGroup()->size());

          basesSaturation.insert(basesSaturation.end(), BaseManagement(baseLocation, baseLocation->getCollectMinerals()->getGroup()->size(), isUnderAttack));
        }
      }
      else
      {
        if (!stopRecruitingProbes)
          baseLocation->executeRecruitProbes();

        const int probesDiff = baseLocation->getNbProbesOnMinerals() - int(std::ceil(probesPerCrystalRatio * double(baseLocation->getLastKnownNbCrystalsLeft())));
        if (probesDiff > 0)
          basesSaturation.insert(basesSaturation.begin(), BaseManagement(baseLocation, probesDiff, isUnderAttack));
        else
          basesSaturation.insert(basesSaturation.end(), BaseManagement(baseLocation, probesDiff, isUnderAttack));
      }
    }
  }

  //
  // Transfer probes to gas or build assimilators
  //
  typedef general_iterator<std::list<BaseManagement>::iterator, std::list<BaseManagement>::reverse_iterator> list_iterator;
  list_iterator it = totalProbesTransferGas >= 0 ? list_iterator(basesSaturation.begin()) : list_iterator(basesSaturation.rbegin());
  list_iterator end = totalProbesTransferGas >= 0 ? list_iterator(basesSaturation.end()) : list_iterator(basesSaturation.rend());
  for (; totalProbesTransferGas != 0 && it != end; ++it)
  {
    BaseManagement& overSaturation = *it;
    if (!overSaturation.isUnderAttack)
    {
      const size_t probesOnGas = overSaturation.baseLocation->getNbProbesOnGas();
      const size_t maxProbesOnGas = overSaturation.baseLocation->probesGasSaturation();
      const size_t toTransferToGas = std::min(std::min(overSaturation.baseLocation->getNbProbesOnMinerals(), maxProbesOnGas - probesOnGas), totalProbesTransferGas);

      const size_t nbGeysersNeeded = size_t(std::ceil((probesOnGas + toTransferToGas) / 3.));
      for (size_t n = overSaturation.baseLocation->getNbOwnUnits(BWAPI::UnitTypes::Protoss_Assimilator); n < nbGeysersNeeded; ++n)
      {
        overSaturation.baseLocation->buildAssimilator();
      }

      const size_t maxCurrentProbesOnGas = 3 * overSaturation.baseLocation->getNbCompletedOwnUnits(BWAPI::UnitTypes::Protoss_Assimilator);
      const size_t toTransferCurrently = std::min(toTransferToGas, maxCurrentProbesOnGas - probesOnGas);
      GroupTacticalAction::transfer(overSaturation.baseLocation->getCollectMinerals(), overSaturation.baseLocation->getCollectGas(), toTransferCurrently);

      overSaturation.probesDiff -= toTransferToGas;

      totalProbesTransferGas -= toTransferToGas;
    }
  }

  //
  // Transfer (minerals) probes to under-saturated bases
  //
  for (std::list<BaseManagement>::iterator it = basesSaturation.begin();
       it != basesSaturation.end();
       ++it)
  {
    BaseManagement& overSaturation = *it;

    if (overSaturation.probesDiff > 1) // > 1 (instead of > 0) to avoid "flickering" probes between bases
    {
      for (std::list<BaseManagement>::reverse_iterator rit = basesSaturation.rbegin();
           overSaturation.probesDiff > 0 && rit != basesSaturation.rend();
           ++rit)
      {
        BaseManagement& underSaturation = *rit;
        if (underSaturation.probesDiff > overSaturation.probesDiff)
          break;
        const int nbToTransfer = std::min(overSaturation.probesDiff, -underSaturation.probesDiff);
        const int nbTransferred = GroupTacticalAction::transfer(overSaturation.baseLocation->getCollectMinerals(), underSaturation.baseLocation->getCollectMinerals(), nbToTransfer);
        overSaturation.probesDiff -= nbTransferred;
        underSaturation.probesDiff += nbTransferred;
      }
    }
  }
}
