#pragma once
#include "PrimitiveLocation.h"
#include "Arbitrary/CircularLocation.h"

class Map;
class Region;

const int BASE_LOCATION_RADIUS = 4*TILE_SIZE;

/**
* @brief Locations where one player may collect minerals or gas.
*/
class BaseLocation : public virtual PrimitiveLocation
{
public:
  template <class LocationTypes>
  BaseLocation(LocationTypes*, const BWTA::BaseLocation* location, Region* parent)
    : PrimitiveLocation(parent)
    , StaticLocation(parent)
    , m_location(location)
  {
#ifdef _DEBUG
    const std::set<BWAPI::TilePosition>& geysersPosition = m_location->getGeysersPosition();
    for (std::set<BWAPI::TilePosition>::const_iterator it = geysersPosition.begin(); it != geysersPosition.end(); ++it)
    {
      const std::set<BWAPI::Unit*>& unitsOnTile = BWAPI::Broodwar->getUnitsOnTile(it->x(), it->y());
      if (unitsOnTile.size() > 0) // For all accessible geysers
      {
        bool found = false;
        for (std::set<BWAPI::Unit*>::const_iterator itUnit = unitsOnTile.begin(); itUnit != unitsOnTile.end(); ++itUnit)
        {
          if ((*unitsOnTile.begin())->getType() == BWAPI::UnitTypes::Resource_Vespene_Geyser)
            found = true;
        }
        assert(found);
      }
    }
#endif
  }

  Region* getParent();
  const Region* getParent() const;

  virtual bool isInLocation(BWAPI::Position position) const
  {
    return CircularLocation(m_location->getPosition(), BASE_LOCATION_RADIUS).isInLocation(position);
  }

  virtual BWAPI::Position getCurrentCenter() const
  {
    return m_location->getPosition();
  }

  bool hasEnemyResourceDepot() const
  {
    if (BWAPI::Broodwar->enemy()->getRace() == BWAPI::Races::Protoss)
    {
      return getNbEnemyUnits(BWAPI::UnitTypes::Protoss_Nexus) > 0;
    }
    else if (BWAPI::Broodwar->enemy()->getRace() == BWAPI::Races::Terran)
    {
      return getNbEnemyUnits(BWAPI::UnitTypes::Terran_Command_Center) > 0;
    }
    else
    {
      assert(BWAPI::Broodwar->enemy()->getRace() == BWAPI::Races::Zerg);
      return getNbEnemyUnits(BWAPI::UnitTypes::Zerg_Hatchery) > 0 ||
             getNbEnemyUnits(BWAPI::UnitTypes::Zerg_Lair) > 0 ||
             getNbEnemyUnits(BWAPI::UnitTypes::Zerg_Hive) > 0;
    }
  }

  size_t getLastKnownNbEnemyRefineries() const
  {
    if (BWAPI::Broodwar->enemy()->getRace() == BWAPI::Races::Protoss)
    {
      return getNbEnemyUnits(BWAPI::UnitTypes::Protoss_Assimilator);
    }
    else if (BWAPI::Broodwar->enemy()->getRace() == BWAPI::Races::Terran)
    {
      return getNbEnemyUnits(BWAPI::UnitTypes::Terran_Refinery);
    }
    else
    {
      assert(BWAPI::Broodwar->enemy()->getRace() == BWAPI::Races::Zerg);
      return getNbEnemyUnits(BWAPI::UnitTypes::Zerg_Extractor);
    }
  }

  typedef neutral_unit_type_const_iterator crystal_iterator;
  std::pair<crystal_iterator, crystal_iterator> getLastKnownCrystalsLeft() const
  {
    return getNeutralUnits(BWAPI::UnitTypes::Resource_Mineral_Field);

    //std::pair<neutral_unit_type_const_iterator, neutral_unit_type_const_iterator> units = getNeutralUnits(BWAPI::UnitTypes::Resource_Mineral_Field);
    //// If ever this becomes false the new minerals must be returned as well
    //assert(getNbNeutralUnits(BWAPI::UnitTypes::Resource_Mineral_Field_Type_2) == 0);
    //assert(getNbNeutralUnits(BWAPI::UnitTypes::Resource_Mineral_Field_Type_3) == 0);

    //return std::make_pair(crystal_iterator(units.first), crystal_iterator(units.second));
  }

  size_t getLastKnownNbCrystalsLeft() const
  {
    return getNbNeutralUnits(BWAPI::UnitTypes::Resource_Mineral_Field);
  }

  //bool isPossibleStartLocation() const
  //{
  //  // BWTA bug?
  //  return m_location->isStartLocation();
  //}

  size_t probesMineralsSaturation() const
  {
    return static_cast<size_t>(std::ceil(2.5 * double(getLastKnownNbCrystalsLeft())));
  }

  size_t probesGasSaturation() const
  {
    return 3 * getNbGeysers();
  }

  size_t getNbGeysers() const
  {
    return m_location->getGeysersPosition().size();
  }

  const BWTA::BaseLocation* getBWTABaseLocation() const
  {
    return m_location;
  }

  virtual void output(std::ostream& out) const
  {
    out << "Base location " << m_location->getPosition();
  }

protected:
  friend class Map;

  const BWTA::BaseLocation* const m_location;
};
