#include "stdafx.h"
#include "PerceptualState.h"
#include "Locations/Common.h"
#include <BWAPI.h>
#include <BWTA.h>
#include <algorithm>
#include <cassert>
#include <cmath>
#include <functional>
#include <limits>
#include <boost/numeric/conversion/bounds.hpp>

static const double MAX_TILES_DISTANCE_BUILDING_FROM_BASE_LOCATION = 100;

double ray_segment_intersection(const Vector2f& point, const double angle, const Vector2f& p1, const Vector2f& p2, Vector2f& intersection);

PerceptualState::PerceptualState()
  : m_enemyInfluenceMap(BWAPI::Broodwar->mapWidth(), BWAPI::Broodwar->mapHeight(), BWAPI::Broodwar->enemy())
  , m_selfInfluenceMap(BWAPI::Broodwar->mapWidth(), BWAPI::Broodwar->mapHeight(), BWAPI::Broodwar->self())
  , m_visibilityInfluenceMap(BWAPI::Broodwar->mapWidth(), BWAPI::Broodwar->mapHeight(), BWAPI::Broodwar->self())
  , m_structureInfluenceMap(BWAPI::Broodwar->mapWidth(), BWAPI::Broodwar->mapHeight(), BWAPI::Broodwar->enemy())
  , m_map(NULL)
  , m_updateEnemyThreatInfluenceMap(*this, m_enemyInfluenceMap)
  , m_updateSelfThreatInfluenceMap(*this, m_selfInfluenceMap)
  , m_updateVisibilityInfluenceMap(*this, m_visibilityInfluenceMap)
  , m_updateStructuralImportanceInfluenceMap(*this, m_structureInfluenceMap)
  , m_updateBuildings(NULL)
  , m_updateUnits(NULL)
{
}

PerceptualState::~PerceptualState()
{
  delete m_updateBuildings;
  delete m_updateUnits;
  delete m_map;
}

void PerceptualState::activateTasks()
{
  Spar->getScheduler().activate(&m_updateEnemyThreatInfluenceMap, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 4, 1);
  Spar->getScheduler().activate(&m_updateSelfThreatInfluenceMap, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 4, 2);
  Spar->getScheduler().activate(&m_updateVisibilityInfluenceMap, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 4, 3);
  Spar->getScheduler().activate(&m_updateStructuralImportanceInfluenceMap, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 4, 4);

  m_updateBuildings = new UpdateBuildingsTask(*this, m_map);
  m_updateUnits = new UpdateUnitsTask(*this, m_map);
  Spar->getScheduler().activate(m_updateBuildings, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 5, 0); // period must be < than nb frames for vision to fade!
  Spar->getScheduler().activate(m_updateUnits, Scheduler::SCHEDULING_MEDIUM_PRIORITY, 12, 0);
}

void PerceptualState::monitorUnitCompletion(BWAPI::Unit* unit)
{
  if(unit->getPlayer() != BWAPI::Broodwar->self())
  {
    //this unit is not ours.
  }
  else
  {
    if (unit->isCompleted())
    {
      if(unit->getType() == BWAPI::UnitTypes::Protoss_Archon || unit->getType() == BWAPI::UnitTypes::Protoss_Dark_Archon)
      {
        std::pair<stdext::hash_map<BWAPI::Unit*, UnitMorphing>::iterator, bool> result =
          m_unitsMorphing.insert(std::make_pair(unit, UnitMorphing(unit)));
        if (result.second)
        {
          UnitMorphing& unitMorphing = result.first->second;
          unitMorphing.setPosInContainer(result.first);
          unitMorphing.execute();
        }
      }
      else
        getUnitCompletedEvent().raise(unit);
    }
    else
    {
      std::pair<stdext::hash_map<BWAPI::Unit*, UnitInConstruction>::iterator, bool> result =
        m_unitsInConstruction.insert(std::make_pair(unit, UnitInConstruction(unit)));
      if (result.second)
      {
        UnitInConstruction& unitInConstruction = result.first->second;
        unitInConstruction.setPosInContainer(result.first);
        unitInConstruction.execute();
      }
    }
  }
}

// Only send onUnitDiscover to primitive locations, since it is the first event received

void PerceptualState::onUnitDiscover(BWAPI::Unit* unit)
{
  PrimitiveLocation* newLocation = getHostLocation(unit);
  assert(newLocation != NULL);

  newLocation->onUnitDiscover(unit);

  monitorUnitCompletion(unit);

  m_unitDiscoverEvent.raise(unit);
}

void PerceptualState::onUnitEvade(BWAPI::Unit* unit)
{
  m_unitEvadeEvent.raise(unit);
}

void PerceptualState::onUnitShow(BWAPI::Unit* unit)
{
  m_unitShowEvent.raise(unit);
}

void PerceptualState::onUnitHide(BWAPI::Unit* unit)
{
  m_unitHideEvent.raise(unit);
}

void PerceptualState::onUnitCreate(BWAPI::Unit* unit)
{
  m_unitCreateEvent.raise(unit);
}

void PerceptualState::onUnitDestroy(BWAPI::Unit* unit)
{
  m_unitDestroyEvent.raise(unit);
}

void PerceptualState::onUnitMorph(BWAPI::Unit* unit)
{
  monitorUnitCompletion(unit);

  m_unitMorphEvent.raise(unit);
}

void PerceptualState::onUnitRenegade(BWAPI::Unit* unit)
{
  m_unitRenegadeEvent.raise(unit);
}

bool PerceptualState::tileInSight(BWAPI::TilePosition tile, BWAPI::UnitType unitType) const
{
  const int tileWidth = unitType.tileWidth();
  const int tileHeight = unitType.tileHeight();

  bool tileInSight = false;
  for (int x = 0; !tileInSight && x < tileWidth; ++x)
    for (int y = 0; !tileInSight && y < tileHeight; ++y)
      tileInSight = BWAPI::Broodwar->isVisible(tile.x() + x, tile.y() + y);

  return tileInSight;
}

bool PerceptualState::positionIsReachable(const BWAPI::Position& current_pos, const BWAPI::Position& destination) const
  {
    //longer for debug
    BWAPI::TilePosition tilePos(destination);
    bool valid = destination.isValid();
    bool hasPath = current_pos.hasPath(destination);
   // bool isWalkable = BWAPI::Broodwar->isWalkable(position.x()/8, position.y()/8);
    std::set<BWAPI::Unit*> onTile = BWAPI::Broodwar->getUnitsOnTile(tilePos.x(), tilePos.y());
    std::set<BWAPI::Unit*> geysers = BWAPI::Broodwar->getGeysers();
    bool occupiedByGeyser = false;
    for(std::set<BWAPI::Unit*>::iterator it = geysers.begin(); it != geysers.end(); ++it)
    {
      if(tilePos.x() >= (*it)->getTilePosition().x() && tilePos.x() < (*it)->getTilePosition().x() + 4 && //geysers are 4 tiles by 2 tiles
        tilePos.y() >= (*it)->getTilePosition().y() && tilePos.y() < (*it)->getTilePosition().y() + 2){
         occupiedByGeyser = true;
         break;
      }
    }

    if(!valid || !hasPath || !onTile.empty() || occupiedByGeyser)
      return false;
    /*
     if(!position.isValid() ||
       !unit->hasPath(destination) ||
       !BWAPI::Broodwar->isWalkable(destination.x()/8, destination.y()/8) ||
       !BWAPI::Broodwar->getUnitsOnTile(destination.x()/TILE_SIZE, destination.y()/TILE_SIZE).empty())
       return false;*/
     return true;
  }

class IsTargeting : public std::unary_function<BWAPI::Unit*, bool>
{
  BWAPI::Unit* m_target;
public:
  IsTargeting(BWAPI::Unit* target) : m_target(target) {}
  bool operator()(BWAPI::Unit* unit)
  {
    return unit->getOrderTarget() == m_target;
  }
};

std::set<BWAPI::Unit*> PerceptualState::getTargetingUnits(BWAPI::Unit* unit) const
{
  std::set<BWAPI::Unit*> targetingUnits;

  const std::set<BWAPI::Unit*>& enemyUnits = BWAPI::Broodwar->enemy()->getUnits();
  std::remove_copy_if(
    enemyUnits.begin(), enemyUnits.end(),
    std::inserter(targetingUnits, targetingUnits.begin()),
    IsTargeting(unit));

  return targetingUnits;
}

bool PerceptualState::isUnitTargeted(BWAPI::Unit* unit) const
{
  return !getTargetingUnits(unit).empty();
}

PrimitiveLocation* PerceptualState::getHostLocation(BWAPI::UnitType unitType, bool isLifted, BWAPI::Position position, BWAPI::TilePosition tilePosition) const
{
  PrimitiveLocation* hostLocation = NULL;
  if (unitType.isResourceDepot() || unitType.isResourceContainer())
  {
    BWTA::BaseLocation* baseLocation = BWTA::getNearestBaseLocation(position);
    if (baseLocation != NULL)
    {
      if (baseLocation->getTilePosition().getDistance(tilePosition) > MAX_TILES_DISTANCE_BUILDING_FROM_BASE_LOCATION)
        hostLocation = NULL;
      else
        hostLocation = m_map->getBaseLocation(baseLocation);
    }
  }
  if (hostLocation == NULL)
  {
    BWTA::Region* bwtaRegion = BWTA::getRegion(tilePosition);
    if (bwtaRegion == NULL)
    {
      // TODO: assert flyer?
      hostLocation = m_map->getAirLocation();
    }
    else
    {
      Region* sparRegion = m_map->getRegion(bwtaRegion);
      hostLocation = sparRegion->getRegionLocation();
    }
  }
  return hostLocation;
}

BWAPI::Position PerceptualState::getGroundFrontier(const BWAPI::Position& position, const double angle) const
{
  BWAPI::Position frontier(position);

  BWTA::Region* region = BWTA::getRegion(BWAPI::TilePosition(position));
  if(region != NULL)
  {
    const BWTA::Polygon& polygon = region->getPolygon();

    if(!polygon.empty())
    {
      double min_dist = boost::numeric::bounds<double>::highest();
      Vector2f intersection(0, 0);

      const Vector2f point(position.x(), position.y());

      // TODO check other polygons (e.g., if intersection lies on a chokepoint)
      for(BWTA::Polygon::const_iterator prev=polygon.begin(), i=prev+1 /*skip first vertex*/;
        i!=polygon.end();
        ++i, ++prev)
      {
        const Vector2f p1(prev->x(), prev->y());
        const Vector2f p2(i->x(), i->y());

        const double distance = ::ray_segment_intersection(point, angle, p1, p2, intersection);
        if(distance >= 0.0 && distance < min_dist)
        {
          frontier.x() = static_cast<int>(intersection.x());
          frontier.y() = static_cast<int>(intersection.y());
          min_dist = distance;
        }
      }

      // check back()---front() segment as well
      {
        const Vector2f first(polygon.front().x(), polygon.front().y());
        const Vector2f last(polygon.back().x(), polygon.back().y());

        const double distance = ::ray_segment_intersection(point, angle, last, first, intersection);
        if(distance >= 0.0 && distance < min_dist)
        {
          frontier.x() = static_cast<int>(intersection.x());
          frontier.y() = static_cast<int>(intersection.y());
          // min_dist = distance;
        }
      }
    }
  }

  return frontier;
}

/**
* @brief Functor that acts as a goal test function for the breadth first search algorithm
*/
struct AtEnemyInfluenceBorder
{
  typedef bool return_type;

  bool operator()(size_t x, size_t y, const InfluenceMap<Threat>& map) const
  {
    const Threat& threat = map.getInfluence(x, y);
    return threat.isNull() != m_current;
  }

  AtEnemyInfluenceBorder(bool current) : m_current(current) {}

private:
  bool m_current;
};

BWAPI::Position PerceptualState::getBorder(const BWAPI::Position& position) const
{
  std::pair<int,int> pos = InfluenceMaps::breadth_first_search(
    position.x() / TILE_SIZE,
    position.y() / TILE_SIZE,
    AtEnemyInfluenceBorder(m_enemyInfluenceMap.getInfluence(position).isNull()), m_enemyInfluenceMap);

  return BWAPI::Position(pos.first * TILE_SIZE, pos.second * TILE_SIZE);
}

struct ThreatGradient
{
  typedef std::pair<float,float> return_type;

  ThreatGradient(bool ground) : m_ground(ground) {}

  return_type operator()(int x, int y, const InfluenceMap<Threat>& selfMap, const InfluenceMap<Threat>& enemyMap)
  {
    ThreatGradientEvaluator evaluator(selfMap, enemyMap, m_ground, x, y);
    BresenhamCircle(evaluator, 10, true);
    return evaluator.getGradient();
  }

private:
  bool m_ground;

  struct ThreatGradientEvaluator : std::binary_function<int,int,void>
  {
    ThreatGradientEvaluator(const InfluenceMap<Threat>& selfMap, const InfluenceMap<Threat>& enemyMap, bool ground, int x, int y)
      : m_self(selfMap), m_enemy(enemyMap), m_ground(ground), m_x(x), m_y(y), m_vx(0.f), m_vy(0.f) {}

    void operator()(int x, int y)
    {
      if (m_self.isValid(x + m_x, y + m_y) && !(x == 0 && y == 0))
      {
        const Threat& threat = m_enemy.getInfluence(x + m_x, y + m_y);
        const Threat& self   = m_self.getInfluence( x + m_x, y + m_y);

        int value = m_ground ?
          threat.ground - self.ground:
        threat.air    - self.air;

        // Penalties
        if (x + m_x == 0 || x + m_x == m_self.getWidth()  - 1) value += 100;
        if (y + m_y == 0 || y + m_y == m_self.getHeight() - 1) value += 100;
        //if (!BWAPI::Broodwar->isWalkable(x + m_x, y + m_y))   value += 100;
        if(!Spar->getPerceptualState().positionIsReachable(BWAPI::Position(x * TILE_SIZE, y * TILE_SIZE), BWAPI::Position((x + m_x) * TILE_SIZE, (y + m_y) * TILE_SIZE)))
          value += 100;

        float norm = std::sqrtf(static_cast<float>(x*x + y*y));

        m_vx += value * x/norm;
        m_vy += value * y/norm;
      }
    }

    return_type getGradient() const
    {
      float norm = std::sqrtf(m_vx*m_vx + m_vy*m_vy);
      return return_type(m_vx/norm, m_vy/norm);
    }

  private:
    const InfluenceMap<Threat>& m_self;
    const InfluenceMap<Threat>& m_enemy;
    bool m_ground;
    int m_x;
    int m_y;

    float m_vx;
    float m_vy;
  };
};

struct StopCriterion
{
  StopCriterion(int max_iterations) : m_max(max_iterations) {}

  bool operator()(const std::pair<float,float>& gradient)
  {
    if (m_max-- <= 0) return true;

    float sqnorm = gradient.first*gradient.first + 
      gradient.second*gradient.second;
    return sqnorm < 1e-3f;
  }

private:
  int m_max;
};

BWAPI::Position PerceptualState::getSafeRetreatPosition(BWAPI::Unit* unit) const
{
  std::pair<int,int> pos = InfluenceMaps::gradient_descent(
    unit->getTilePosition().x(),
    unit->getTilePosition().y(),
    ThreatGradient(!unit->getType().isFlyer()),
    StopCriterion(5),
    m_selfInfluenceMap,
    m_enemyInfluenceMap);

  SPAR_LOG(LogTypes::PERCEPTUAL_STATE, "Start: (%d,%d)\tDestination: (%d,%d)", unit->getTilePosition().x(), unit->getTilePosition().y(), pos.first, pos.second);
  return BWAPI::Position(pos.first * TILE_SIZE, pos.second * TILE_SIZE);
}

BWAPI::Position PerceptualState::getSafeDodgePosition(BWAPI::Unit* unit) const
{
  BWAPI::TilePosition unitPosition = BWAPI::TilePosition(unit->getPosition());
  int x, y;
  BWAPI::TilePosition bestPosition;
  Threat currentBestInfluence;
  currentBestInfluence.ground = boost::numeric::bounds<int>::lowest();
  currentBestInfluence.air = boost::numeric::bounds<int>::lowest();
  for(int i= -SAFE_POSITION_RANGE; i < SAFE_POSITION_RANGE + 1; ++i)
  {
    for(int j= -SAFE_POSITION_RANGE; j < SAFE_POSITION_RANGE + 1; ++j)
    {
      x= unitPosition.x() + i;
      y= unitPosition.y() + j;
      if (m_selfInfluenceMap.isValid(x, y))
      {
        int groundDiff = m_selfInfluenceMap.getInfluence(x, y).ground - m_enemyInfluenceMap.getInfluence(x, y).ground;
        int airDiff = m_selfInfluenceMap.getInfluence(x, y).air - m_enemyInfluenceMap.getInfluence(x, y).air;
        if ((unit->getType().isFlyer() && airDiff >= currentBestInfluence.air) 
          || (!unit->getType().isFlyer() && groundDiff >= currentBestInfluence.ground))
        {
          currentBestInfluence.ground = groundDiff;
          currentBestInfluence.air = airDiff;
          bestPosition = BWAPI::TilePosition(BWAPI::TilePosition(x, y));
        }
      }
    }
  }

  BWAPI::Position destination = BWAPI::Position(bestPosition) - unit->getPosition();
  float length = sqrt((float)(destination.x() * destination.x() + destination.y() * destination.y()));
  float newX = destination.x() / length; 
  float newY = destination.y() / length;
  float distance = DODGE_DISTANCE;
  destination = unit->getPosition() + BWAPI::Position(static_cast<int>(newX * distance), 
                                                      static_cast<int>(newY * distance));

  return destination;
}

BWAPI::Position PerceptualState::getSafeRetreatPosition(const PersistentUnitGroup* group, const UnitGroup& enemyUnits) const
{
  // TODO: delete eventually! retreat positions should be assigned to clusters of units
  //       instead of entire group!

  BWAPI::Position enemiesVector(0, 0);
  for (UnitGroup::const_iterator it = enemyUnits.begin();
    it != enemyUnits.end();
    ++it)
  {
    for (PersistentUnitGroup::const_iterator itUnit = group->begin();
      itUnit != group->end();
      ++itUnit)
    {
      enemiesVector += ((*it)->getPosition() - (*itUnit)->getPosition());
    }
  }
  const double destinationAngle = atan2(static_cast<double>(-enemiesVector.y()),
    static_cast<double>(-enemiesVector.x()));
  BWAPI::Position center = group->getCenter();
  BWAPI::Position destination = this->getGroundFrontier(center, destinationAngle);

  return destination;
}

struct MaxUnvisitedTile
{
  MaxUnvisitedTile(const BWTA::Region* const region)
    :m_max(0L) 
    ,m_region(region)
    ,m_x(-1)
    ,m_y(-1)
    //,m_unit(unit)
  {}

  void operator()(size_t x, size_t y, const InfluenceMap<Visibility>& map)
  {
    const Visibility& visibility = map.getInfluence(x, y);
    BWAPI::Position position(x*TILE_SIZE, y*TILE_SIZE);
    if (m_region == BWTA::getRegion(x, y))
    {
      if (visibility.lastVisited > m_max)
      {
        m_max = visibility.lastVisited;
        m_x = x;
        m_y = y;
      }
    }
  }

  BWAPI::Position getMax() const
  {
    return (m_x == -1 || m_y == -1) ? BWAPI::Positions::Invalid : BWAPI::Position(m_x*TILE_SIZE, m_y*TILE_SIZE);
  }

private:
  unsigned long m_max;
  const BWTA::Region* const m_region;
  int m_x, m_y;
  BWAPI::Unit* m_unit;
};

//probablement pu besoin du unit... still in testing
BWAPI::Position PerceptualState::getScoutingLocation(BWAPI::Unit* unit, const BWTA::Region* const region) const
{
  return m_visibilityInfluenceMap.for_each(MaxUnvisitedTile(region)).getMax();
}

///////////////////////////
// UnitInConstruction
///////////////////////////

PerceptualState::UnitInConstruction::UnitInConstruction(BWAPI::Unit* unit)
: m_unitInConstruction(unit)
, m_timeoutEvent(*this, Scheduler::SCHEDULING_MEDIUM_PRIORITY,1,0)
, m_unitDestroyedEventHandler(*this)
{
}

PerceptualState::UnitInConstruction::UnitInConstruction(const PerceptualState::UnitInConstruction& other)
: m_unitInConstruction(other.m_unitInConstruction)
, m_timeoutEvent(*this, Scheduler::SCHEDULING_MEDIUM_PRIORITY,1,0)
, m_unitDestroyedEventHandler(*this)
{
}

void PerceptualState::UnitInConstruction::execute()
{
  assert(!m_unitInConstruction->isCompleted());
  assert(m_unitInConstruction->isBeingConstructed());

  // It seems to take one additionnal frame to complete the unit (Protoss buildings take 71 frame to be useable after the remainingBuildTime hits zero)
  if(m_unitInConstruction->getType().isBuilding())
  {
    m_timeoutEvent.setWaitingTimeBeforeExecution(m_unitInConstruction->getRemainingBuildTime()+71);
  }else
  {
    m_timeoutEvent.setWaitingTimeBeforeExecution(m_unitInConstruction->getRemainingBuildTime()+1);
  }
  m_timeoutEvent.activate();
  m_unitDestroyedSubscriberId = Spar->getUnitManager().subscribeIndividualUnitDestroy(m_unitInConstruction, &m_unitDestroyedEventHandler, NULL);
}

void PerceptualState::UnitInConstruction::onTimeout()
{
  assert(m_unitInConstruction->exists());
  if(m_unitInConstruction->isCompleted() || !m_unitInConstruction->isBeingConstructed())
  {
    assert(m_unitInConstruction->isCompleted());
    assert(!m_unitInConstruction->isBeingConstructed());

    SPAR_LOG(LogTypes::PERCEPTUAL_STATE, "Unit completed (0x%x) %s %d"
      , m_unitInConstruction
      , m_unitInConstruction->getType().getName().c_str()
      , m_unitInConstruction->getID());
    Spar->getPerceptualState().getUnitCompletedEvent().raise(m_unitInConstruction);
    terminate();
  } else {
    SPAR_LOG(LogTypes::GENERAL_ERROR, "Unit not completed when it should");
    m_timeoutEvent.deactivate();
    // +1 to make sure we do not start an infinite loop
    m_timeoutEvent.setWaitingTimeBeforeExecution(m_unitInConstruction->getRemainingBuildTime()+1);
    m_timeoutEvent.activate();
  }
}

void PerceptualState::UnitInConstruction::onUnitDestroyed(void* data)
{
  terminate();
}

void PerceptualState::UnitInConstruction::terminate()
{
  m_timeoutEvent.deactivate();
  Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_unitInConstruction, m_unitDestroyedSubscriberId);
  Spar->getPerceptualState().onUnitInConstructionTerminate(this);
}

///////////////////////////
// UnitMorphing
///////////////////////////

PerceptualState::UnitMorphing::UnitMorphing(BWAPI::Unit* unit)
: m_unitMorphing(unit)
, m_timeoutEvent(*this, Scheduler::SCHEDULING_MEDIUM_PRIORITY,1,0)
, m_unitDestroyedEventHandler(*this)
{
}

PerceptualState::UnitMorphing::UnitMorphing(const PerceptualState::UnitMorphing& other)
: m_unitMorphing(other.m_unitMorphing)
, m_timeoutEvent(*this, Scheduler::SCHEDULING_MEDIUM_PRIORITY,1,0)
, m_unitDestroyedEventHandler(*this)
{
}

void PerceptualState::UnitMorphing::execute()
{
  assert(!m_unitMorphing->isMorphing());

  m_timeoutEvent.setWaitingTimeBeforeExecution(m_unitMorphing->getRemainingBuildTime());

  m_timeoutEvent.activate();
  m_unitDestroyedSubscriberId = Spar->getUnitManager().subscribeIndividualUnitDestroy(m_unitMorphing, &m_unitDestroyedEventHandler, NULL);
}

void PerceptualState::UnitMorphing::onTimeout()
{
  assert(m_unitMorphing->exists());
  if(!m_unitMorphing->isMorphing())
  {
    SPAR_LOG(LogTypes::PERCEPTUAL_STATE, "Unit morph terminated (0x%x) %s %d"
      , m_unitMorphing
      , m_unitMorphing->getType().getName().c_str()
      , m_unitMorphing->getID());
    Spar->getPerceptualState().getUnitMorphCompletedEvent().raise(m_unitMorphing);
    terminate();
  } else {
    SPAR_LOG(LogTypes::GENERAL_ERROR, "Unit not morphed when it should");
    m_timeoutEvent.deactivate();
    // +1 to make sure we do not start an infinite loop
    m_timeoutEvent.setWaitingTimeBeforeExecution(m_unitMorphing->getRemainingBuildTime()+1);
    m_timeoutEvent.activate();
  }
}

void PerceptualState::UnitMorphing::onUnitDestroyed(void* data)
{
  terminate();
}

void PerceptualState::UnitMorphing::terminate()
{
  m_timeoutEvent.deactivate();
  Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_unitMorphing, m_unitDestroyedSubscriberId);
  Spar->getPerceptualState().onUnitMorphingTerminate(this);
}

///////////////////////////
// ResearchInProgress
///////////////////////////
PerceptualState::ResearchInProgress* PerceptualState::onResearchStarted(const Process* parent, TechTypeTech* tech, BWAPI::Unit* building)
{
  ResearchInProgress* research = new ResearchInProgress(parent, building, tech);
  Process::zombifyOnTerminate(research);
  research->execute();
  return research;
}

PerceptualState::ResearchInProgress* PerceptualState::onUpgradeStarted(const Process* parent, UpgradeTypeTech* upgrade, BWAPI::Unit* building)
{
  ResearchInProgress* research = new ResearchInProgress(parent, building, upgrade);
  Process::zombifyOnTerminate(research);
  research->execute();
  return research;
}

const Process::TerminationMode PerceptualState::ResearchInProgress::ResearchCancelled;
const Process::TerminationMode PerceptualState::ResearchInProgress::BuildingDestroyed;

PerceptualState::ResearchInProgress::ResearchInProgress(const Process* parent, BWAPI::Unit* unit, TechTypeTech* tech)
  : Process(parent)
  , m_building(unit)
  , m_tech(tech)
  , m_timeoutEvent(*this, Scheduler::SCHEDULING_MEDIUM_PRIORITY,1,0)
  , m_unitDestroyedEventHandler(*this)
  , m_upgrade(NULL)
{
}

PerceptualState::ResearchInProgress::ResearchInProgress(const Process* parent, BWAPI::Unit* unit, UpgradeTypeTech* upgrade)
  : Process(parent)
  , m_building(unit)
  , m_upgrade(upgrade)
  , m_timeoutEvent(*this, Scheduler::SCHEDULING_MEDIUM_PRIORITY,1,0)
  , m_unitDestroyedEventHandler(*this)
  , m_tech(NULL)
{
}

void PerceptualState::ResearchInProgress::output(std::ostream& out) const
{
  out << "Research in progress: ";
  if (m_tech != NULL)
    out << *m_tech;
  else
    out << *m_upgrade;
}

void PerceptualState::ResearchInProgress::executeImpl()
{
  if(m_tech != NULL)
  {
    m_timeoutEvent.setWaitingTimeBeforeExecution(m_building->getRemainingResearchTime()+3);
  }
  else
  {
    m_timeoutEvent.setWaitingTimeBeforeExecution(m_building->getRemainingUpgradeTime()+3);
  }
  m_timeoutEvent.activate();
  m_unitDestroyedSubscriberId = Spar->getUnitManager().subscribeIndividualUnitDestroy(m_building, &m_unitDestroyedEventHandler, NULL);
}

void PerceptualState::ResearchInProgress::onTimeout()
{
  if (m_tech != NULL)
  {
    if (BWAPI::Broodwar->self()->hasResearched(m_tech->getTech()))
    {
      m_tech->setTechAcquired();
      terminate(Success);
    }
    else if (!m_building->isResearching() || m_building->getRemainingResearchTime() == 0)
    {
      terminate(ResearchCancelled);
    }
    else
    {
      SPAR_LOG(LogTypes::GENERAL_ERROR, "Research not ready when it should: %s", getObjectDescription(m_tech).c_str());
      m_timeoutEvent.deactivate();
      // +1 to make sure we do not start an infinite loop
      m_timeoutEvent.setWaitingTimeBeforeExecution(m_building->getRemainingResearchTime()+1);
      m_timeoutEvent.activate();
    }
  }
  else
  {
    if (BWAPI::Broodwar->self()->getUpgradeLevel(m_upgrade->getUpgrade()) >= m_upgrade->getLevel())
    {
      m_upgrade->setTechAcquired();
      terminate(Success);
    }
    else if (!m_building->isUpgrading() || m_building->getRemainingUpgradeTime() == 0)
    {
      terminate(ResearchCancelled);
    }
    else
    {
      SPAR_LOG(LogTypes::GENERAL_ERROR, "Research not ready when it should: %s", getObjectDescription(m_upgrade).c_str());
      m_timeoutEvent.deactivate();
      // +1 to make sure we do not start an infinite loop
      m_timeoutEvent.setWaitingTimeBeforeExecution(m_building->getRemainingUpgradeTime()+1);
      m_timeoutEvent.activate();
    }
  }
}

void PerceptualState::ResearchInProgress::onUnitDestroyed(void* data)
{
  terminate(BuildingDestroyed);
}

void PerceptualState::ResearchInProgress::cancel()
{
  if (m_tech != NULL)
    m_building->cancelResearch();
  else
    m_building->cancelUpgrade();
  terminate(ResearchCancelled);
}

void PerceptualState::ResearchInProgress::terminateImpl()
{
  m_timeoutEvent.deactivate();
  Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_building, m_unitDestroyedSubscriberId);
}

void PerceptualState::displayMap()
{
  //we will iterate through all the base locations, and draw their outlines.
  for(std::set<BWTA::BaseLocation*>::const_iterator i=BWTA::getBaseLocations().begin();i!=BWTA::getBaseLocations().end();i++)
  {
    BWAPI::TilePosition p=(*i)->getTilePosition();
    BWAPI::Position c=(*i)->getPosition();

    //draw outline of center location
    BWAPI::Broodwar->drawBox(BWAPI::CoordinateType::Map,p.x()*32,p.y()*32,p.x()*32+4*32,p.y()*32+3*32,BWAPI::Colors::Blue,false);

    //draw a circle at each mineral patch
    for(std::set<BWAPI::Unit*>::const_iterator j=(*i)->getStaticMinerals().begin();j!=(*i)->getStaticMinerals().end();j++)
    {
      BWAPI::Position q=(*j)->getInitialPosition();
      BWAPI::Broodwar->drawCircle(BWAPI::CoordinateType::Map,q.x(),q.y(),30,BWAPI::Colors::Cyan,false);
    }

    //draw the outlines of vespene geysers
    for(std::set<BWAPI::TilePosition>::const_iterator j=(*i)->getGeysersPosition().begin();j!=(*i)->getGeysersPosition().end();j++)
    {
      BWAPI::TilePosition q=*j;
      BWAPI::Broodwar->drawBox(BWAPI::CoordinateType::Map,q.x()*32,q.y()*32,q.x()*32+4*32,q.y()*32+2*32,BWAPI::Colors::Orange,false);
    }

    //if this is an island expansion, draw a yellow circle around the base location
    if ((*i)->isIsland())
    {
      BWAPI::Broodwar->drawCircle(BWAPI::CoordinateType::Map,c.x(),c.y(),80,BWAPI::Colors::Yellow,false);
    }
  }

  //we will iterate through all the regions and draw the polygon outline of it in green.
  for(std::set<BWTA::Region*>::const_iterator r=BWTA::getRegions().begin();r!=BWTA::getRegions().end();r++)
  {
    BWTA::Polygon p=(*r)->getPolygon();
    for(int j=0;j<(int)p.size();j++)
    {
      BWAPI::Position point1=p[j];
      BWAPI::Position point2=p[(j+1) % p.size()];
      BWAPI::Broodwar->drawLine(BWAPI::CoordinateType::Map,point1.x(),point1.y(),point2.x(),point2.y(),BWAPI::Colors::Green);
    }
  }

  //we will visualize the chokepoints with red lines
  for(std::set<BWTA::Region*>::const_iterator r=BWTA::getRegions().begin();r!=BWTA::getRegions().end();r++)
  {
    for(std::set<BWTA::Chokepoint*>::const_iterator c=(*r)->getChokepoints().begin();c!=(*r)->getChokepoints().end();c++)
    {
      BWAPI::Position point1=(*c)->getSides().first;
      BWAPI::Position point2=(*c)->getSides().second;
      BWAPI::Broodwar->drawLine(BWAPI::CoordinateType::Map,point1.x(),point1.y(),point2.x(),point2.y(),BWAPI::Colors::Red);
    }
  }
}

double ray_segment_intersection(const Vector2f& point, const double angle, const Vector2f& p1, const Vector2f& p2, Vector2f& intersection)
{
  const Vector2f segment = p2 - p1;
  const Vector2f ray(cos(angle), sin(angle));

  const Vector2f psegment = segment.perpendicular();
  const Vector2f pray = ray.perpendicular();

  // Make sure that lines are not parallel
  const float d = psegment.dot(ray);
  if(d != 0.f)
  {
    const float g = psegment.dot(p1 - point) / d;
    const float h = pray.dot(p1 - point) / d;

    if(g >= 0 && 0 <= h && h <= 1)
    {
      intersection = point + g * ray;
      const double distance = g; // ray is a unit-length vector

      //#ifdef _DEBUG
      //        Vector2f intersection2 = p1 + h*segment;
      //        double distance2 = point.distance_to(intersection);
      //
      //        assert(intersection.almost_equal_to(intersection2));
      //        assert(fabs(distance - distance2) < 1e-3);
      //#endif

      return distance;
    }
  }

  return -1.0;
}
