#include "stdafx.h"
#include "Tech.h"
#include "Techs.h"

void Tech::addDependency(BWAPI::UnitType unitType)
{
  std::cout << "  Add dependency \"" << unitType.getName() << "\"" << std::endl;

  Tech* dependency = Techs::getUnitTypeTech(unitType);
  assert(dependency != NULL);
  m_incomingDependencies.insert(dependency);
  dependency->addOutgoingDependency(this);
}

void Tech::addDependency(BWAPI::TechType techType)
{
  std::cout << "  Add dependency \"" << techType.getName() << "\"" << std::endl;
  Tech* dependency = Techs::getTechTypeTech(techType);
  assert(dependency != NULL);
  m_incomingDependencies.insert(dependency);
  dependency->addOutgoingDependency(this);
}

void Tech::addDependency(BWAPI::UpgradeType upgradeType, int level)
{
  std::cout << "  Add dependency \"" << upgradeType.getName() << "\"" << " level " << level << std::endl;
  Tech* dependency = Techs::getUpgradeTypeTech(upgradeType, level);
  assert(dependency != NULL);
  m_incomingDependencies.insert(dependency);
  dependency->addOutgoingDependency(this);
}

bool Tech::allDependenciesAcquired() const
{ 
  bool allDependenciesAcquired = true;
  for (std::set<Tech*>::const_iterator it = m_incomingDependencies.begin(); allDependenciesAcquired && it != m_incomingDependencies.end(); ++it)
  {
    const Tech* dependency = *it;
    if (!dependency->isAcquired())
      allDependenciesAcquired = false;
  }
  return allDependenciesAcquired;
}

void Tech::setTechAcquired()
{
  assert(!m_isAcquired);

  SPAR_LOG(LogTypes::PERCEPTUAL_STATE, "Tech acquired: %s", getObjectDescription(this).c_str());

  m_isAcquired = true;
  Techs::getTechAcquiredEvent().raise(this);

  for (std::set<Tech*>::iterator it = m_outgoingDependencies.begin(); it != m_outgoingDependencies.end(); ++it)
  {
    Tech* outgoingDependency = *it;
    outgoingDependency->onIncomingDependencyAcquired(this);
  }

  assert(m_isAcquired);
  assert(isAcquired());
}

UnitTypeTech::UnitTypeTech(BWAPI::UnitType unitType)
  : m_unitType(unitType)
{
  m_name = unitType.getName();
  std::cout << "Initializing \"" << *this << "\"" << std::endl;

  Techs::addUnitTypeTech(this);

  const std::map<BWAPI::UnitType, int>& requiredUnits = m_unitType.requiredUnits();
  for (std::map<BWAPI::UnitType, int>::const_iterator it = requiredUnits.begin(); it != requiredUnits.end(); ++it)
  {
    addDependency(it->first);
  }

  BWAPI::TechType requiredTech = m_unitType.requiredTech();
  if (requiredTech != BWAPI::TechTypes::None)
  {
    Tech::addDependency(requiredTech);
  }

  const std::pair<BWAPI::UnitType, int> whatBuilds = m_unitType.whatBuilds();
  if (whatBuilds.first != BWAPI::UnitTypes::None)
  {
    addDependency(whatBuilds.first);
  }

  m_isAcquired = isAcquiredImpl();
}

void UnitTypeTech::addDependency(BWAPI::UnitType unitType)
{
  if (!m_unitType.isBuilding() || unitType.isBuilding())
  {
    Tech::addDependency(unitType);
  }
}

bool UnitTypeTech::isAcquiredImpl() const
{
  if (m_unitType.isBuilding())
    return Techs::s_map->getNbCompletedOwnUnits(m_unitType) > 0;
  else
    return allDependenciesAcquired();
}

void UnitTypeTech::setTechLost()
{
  assert(m_isAcquired);

  SPAR_LOG(LogTypes::PERCEPTUAL_STATE, "Tech lost: %s", getObjectDescription(this).c_str());

  m_isAcquired = false;
  Techs::getTechLostEvent().raise(this);

  for (std::set<Tech*>::iterator it = m_outgoingDependencies.begin(); it != m_outgoingDependencies.end(); ++it)
  {
    Tech* outgoingDependency = *it;
    outgoingDependency->onIncomingDependencyLost(this);
  }

  assert(!m_isAcquired);
  assert(!isAcquired());
}

void UnitTypeTech::onIncomingDependencyAcquired(Tech* dependency)
{
  if (!m_unitType.isBuilding())
  {
    if (allDependenciesAcquired())
      setTechAcquired();
  }
}

void UnitTypeTech::onIncomingDependencyLost(Tech* dependency)
{
  if (!m_unitType.isBuilding() && m_isAcquired)
  {
    setTechLost();
  }
}

TechTypeTech::TechTypeTech(BWAPI::TechType techType)
  : m_techType(techType)
{
  m_name = techType.getName();
  std::cout << "Initializing \"" << *this << "\"" << std::endl;

  Techs::addTechTypeTech(this);

  BWAPI::UnitType whatResearches = m_techType.whatResearches();
  if (whatResearches != BWAPI::UnitTypes::None)
  {
    addDependency(whatResearches);
  }

  m_isAcquired = isAcquiredImpl();
}

bool TechTypeTech::isAcquiredImpl() const
{
  return BWAPI::Broodwar->self()->hasResearched(m_techType);
}

UpgradeTypeTech::UpgradeTypeTech(BWAPI::UpgradeType upgradeType, int level)
  : m_upgradeType(upgradeType)
  , m_level(level)
{
  std::stringstream ss;
  ss << upgradeType.getName() << " level " << level;
  m_name = ss.str();
  std::cout << "Initializing \"" << *this << "\"" << std::endl;

  Techs::addUpgradeTypeTech(this);

  if (level != 1)
    addDependency(m_upgradeType, level - 1);

  BWAPI::UnitType whatUpgrades = m_upgradeType.whatUpgrades();
  if (whatUpgrades != BWAPI::UnitTypes::None)
  {
    addDependency(whatUpgrades);
  }

  BWAPI::UnitType whatsRequired = m_upgradeType.whatsRequired(m_level);
  if (whatsRequired != BWAPI::UnitTypes::None)
  {
    addDependency(whatsRequired);
  }

  m_isAcquired = isAcquiredImpl();
}

UpgradeTypeTech::UpgradeTypeTech(BWAPI::UpgradeType upgradeType)
  : m_upgradeType(upgradeType)
  , m_level(1)
{
  m_name = upgradeType.getName();

  Techs::addUpgradeTypeTech(this);

  BWAPI::UnitType whatUpgrades = m_upgradeType.whatUpgrades();
  if (whatUpgrades != BWAPI::UnitTypes::None)
  {
    addDependency(whatUpgrades);
  }

  BWAPI::UnitType whatsRequired = m_upgradeType.whatsRequired(m_level);
  if (whatsRequired != BWAPI::UnitTypes::None)
  {
    addDependency(whatsRequired);
  }
}

bool UpgradeTypeTech::isAcquiredImpl() const
{
  return BWAPI::Broodwar->self()->getUpgradeLevel(m_upgradeType) >= m_level;
}