#pragma once
#include <BWAPI.h>
#include <iostream>

class PrimitiveLocation;
class CompositeLocation;

/**
 * @brief Units information stored in locations.
 */
class UnitLocationData : public Component
{
public:
  UnitLocationData(BWAPI::Unit* unit, PrimitiveLocation* location)
    : m_unit(unit)
    , m_type(unit->getType())
    , m_player(unit->getPlayer())
    , m_location(location)
    , m_primitiveLocationData(NULL)
    , m_regionData(NULL)
    , m_mapData(NULL)
  {
  }
  virtual ~UnitLocationData() {}
  const BWAPI::Player* getPlayer() const { return m_player; }
  BWAPI::UnitType getType() const { return m_type; }
  PrimitiveLocation* getLocation() const { return m_location; }
  void updateLocation(PrimitiveLocation* newLocation) { m_location = newLocation; }
protected:
  // ADD AN EVENT MECHANISM IF EVER THERE MAY BE MORE THAN THESE THREE SUBSCRIBERS!
  friend class PrimitiveLocation;
  friend class CompositeLocation;
  void* m_primitiveLocationData;
  void* m_regionData;
  void* m_mapData;

  BWAPI::Unit* const m_unit;
  const BWAPI::Player* const m_player;
  BWAPI::UnitType m_type;
  /**
   * @brief The last updated primitive location of this unit.
   */
  PrimitiveLocation* m_location;
};

/**
 * @brief Foreign units are neutral or enemy units.
 */
class ForeignUnit : public UnitLocationData
{
public:
  /**
   * @brief Information saved about a particular unit, before said information becomes unavailable.
   */
  class SavedState
  {
  public:
    SavedState(BWAPI::Unit* unit)
      : m_unitType(unit->getType())
      , m_position(unit->getPosition())
      , m_hitPoints(unit->getHitPoints())
      , m_tilePosition(unit->getTilePosition())
      , m_isLifted(unit->isLifted())
    {}

    BWAPI::UnitType getType() const { return m_unitType; }
    BWAPI::Position getPosition() const { return m_position; }
    BWAPI::TilePosition getTilePosition() const { return m_tilePosition; }
    int getHitPoints() const { return m_hitPoints; }
    bool isLifted() const { return m_isLifted; }

  private:
    // Add stuff here as necessary
    BWAPI::UnitType m_unitType;
    BWAPI::Position m_position;
    BWAPI::TilePosition m_tilePosition;
    int m_hitPoints;
    bool m_isLifted;
  };

  ForeignUnit(BWAPI::Unit* unitPtr, PrimitiveLocation* location)
    : UnitLocationData(unitPtr, location)
    , m_lastKnownState(unitPtr)
    , m_lastUnitEvadedFrame(-1)
  {}

  /**
   * @return Whether this unit is accessible, i.e. whether it is in our field of view.
   */
  bool isAccessible() const
  {
    // Not yet evaded
    return m_lastUnitEvadedFrame == -1;
  }

  int getLastEvadedFrame() const
  {
    return m_lastUnitEvadedFrame;
  }

  BWAPI::UnitType getType() const { return isAccessible() ? m_unit->getType() : m_lastKnownState.getType(); }
  BWAPI::UnitType getLastType() const { return m_lastKnownState.getType(); }
  BWAPI::TilePosition getTilePosition() const { return isAccessible() ? m_unit->getTilePosition() : m_lastKnownState.getTilePosition(); }
  BWAPI::Position getPosition() const { return isAccessible() ? m_unit->getPosition() : m_lastKnownState.getPosition(); }

  /**
   * @brief Returns the unit pointer associated with this unit. The unit pointer CANNOT BE RETRIEVED IF THE UNIT IS NOT ACCESSIBLE.
   * @return The unit pointer associated with this unit.
   */
  BWAPI::Unit* getUnit() const
  {
    // WARNING: IF THE UNIT IS NOT ACCESSIBLE (IN FOG OF WAR) THE UNIT POINTER MAY NOT BE VALID ANY MORE

    // If unit is not accessible, use saved state instead.
    // However this can be tolerated for minerals since they always have the same pointer.
    assert(isAccessible() || m_lastKnownState.getType().isMineralField());

    return m_unit;
  }

  /**
   * @brief Returns the last saved state associated with this unit. The saved state CANNOT BE RETRIEVED IF THE UNIT IS ACCESSIBLE
   *        (as in this case you should unit the unit pointer's data directly).
   * @return The saved state associated with this unit.
   */
  const SavedState& getLastKnownState() const
  {
    // If unit is accessible, use unit pointer directly
    assert(!isAccessible());

    return m_lastKnownState;
  }

  /**
   * @brief Signals that the unit pointer is not accessible.
   */
  void markAsAccessible()
  {
    m_lastUnitEvadedFrame = -1;
  }

  /**
   * @brief Signals that the unit is transitioning outside the field of view; save data.
   */
  void markAsInaccessible()
  {
    m_lastKnownState = SavedState(m_unit);

    m_lastUnitEvadedFrame = BWAPI::Broodwar->getFrameCount();
  }

protected:
  friend class PrimitiveLocation;

  /**
   * @brief Last frame where the unit was accessible, or -1 if it is currently accessible.
   */
  int m_lastUnitEvadedFrame;
  /**
   * @brief Last saved state of the unit, if the unit is not currently accessible.
   */
  SavedState m_lastKnownState;
};
