#pragma once
#include "../Utils/ReferenceCounted.h"
#include "Utils/Component.h"
#include "Utils/EventHandler.h"
#include "Utils/Event.h"
#include "Utils/Utils.h"
#include "Utils/GroupBase.h"
#include "SparAIContainer.h"
#include "UnitManager.h"
#include <BWAPI.h>
#include <boost/iterator/transform_iterator.hpp>
#include <utility>
#include <hash_set>
#include <set>

/**
 * @brief A persistent unit group keeps track of its units by removing them when they are destroyed.
 *        This is particularly important since we don't want dangling destroyed unit pointers everywhere.
 *        This class is also reference-counted, so that instances are automatically deleted when they are
 *        not referenced any more.
 */
class PersistentUnitGroup : public ReferenceCounted, public Component, public GroupBase<PersistentUnitGroup>
{
protected:
  class UnitWrapper;
  //typedef std::set<UnitWrapper> UnitWrapperSet;
  typedef stdext::hash_set<UnitWrapper, UnitWrapperBaseHashCompare> UnitWrapperSet;
  /**
   * @brief Stores additional info associated to units.
   */
  class UnitWrapper : public UnitWrapperBase
  {
  public:
    UnitWrapper(BWAPI::Unit* unit)
      : UnitWrapperBase(unit) {}
  protected:
    friend class PersistentUnitGroup;
    /**
     * @brief Subscription ID for UnitManager's OnIndividualUnitDestroy event.
     */
    mutable UnitManager::OnIndividualUnitDestroy::SubscriberID m_unitDestroyID;
    /**
     * @brief Position of this unit wrapper in the set of wrappers.
     */
    UnitWrapperSet::const_iterator m_unitPosInSet;
  };
public:
  //typedef boost::transform_iterator<extract_unit, UnitWrapperSet::const_iterator> const_iterator;
  //typedef boost::transform_iterator<extract_unit, UnitWrapperSet::iterator> iterator;
  typedef UnitWrapperSet::const_iterator const_iterator;
  typedef UnitWrapperSet::iterator iterator;
  
  /**
   * @brief Constructs an empty persistent unit group.
   * @param name The debug name of this group.
   */
  PersistentUnitGroup(const std::string& name)
    : m_name(name)
    , m_unitDestroyEventHandler(*this)
  {
  }

  /**
   * @brief Constructs a persistent unit group with the specified sequence of units.
   * @tparam iterator The type of sequence iterator.
   * @param name The debug name of this group.
   * @param begin The beginning of the sequence of units.
   * @param end past The end of the sequence of units.
   */
  template <class iterator>
  PersistentUnitGroup(const std::string& name, iterator begin, iterator end)
    : m_name(name)
    , m_unitDestroyEventHandler(*this)
  {
    for (; begin != end; ++begin)
      insert(*begin);
  }

  /**
   * @brief The persistent unit group's destructor removes the association with its units.
   */
  ~PersistentUnitGroup();

  /**
   * @brief Adds a new unit to this persistent unit group. Subscribes to the new unit's onUnitDestroy event
   *        so that it can be removed from the group upon its deletion.
   * @param unit The unit to add to the group.
   * @return True if the unit wasn't already in the group.
   */
  bool insert(BWAPI::Unit* unit)
  {
    std::pair<UnitWrapperSet::iterator, bool> p = m_group.insert(UnitWrapper(unit));
    if (p.second)
    {
      p.first->m_unitPosInSet = p.first;
      UnitManager::OnIndividualUnitDestroy::SubscriberID unitManagerID = 
        Spar->getUnitManager().subscribeIndividualUnitDestroy(unit, &m_unitDestroyEventHandler, &(p.first->m_unitPosInSet));
      p.first->m_unitDestroyID = unitManagerID;

      m_unitAddedEvent.raise(unit);
    }
    return p.second;
  }

  bool insert(const_iterator hint, BWAPI::Unit* unit)
  {
    UnitWrapperSet::iterator it = m_group.insert(hint, UnitWrapper(unit));

    it->m_unitPosInSet = it;
    UnitManager::OnIndividualUnitDestroy::SubscriberID unitManagerID = 
      Spar->getUnitManager().subscribeIndividualUnitDestroy(unit, &m_unitDestroyEventHandler, &(it->m_unitPosInSet));
    it->m_unitDestroyID = unitManagerID;

    m_unitAddedEvent.raise(unit);
    
    return true;
  }

  /**
   * @brief Adds a sequence of units to this persistent unit group. Subscribes to the new units onUnitDestroy events
   *        so that they can be removed from the group upon their deletion.
   * @param begin The first element to add.
   * @param end The last element, this element will not be added.
   * @return The number of unit added.
   */
  template <class iterator>
  size_t insert(iterator begin, iterator end)
  {
    size_t nb = 0; 
    for (; begin != end; ++begin)
    {
      if(insert(*begin))
        ++nb;
    }
    return nb;
  }

  /**
   * @brief Removes a unit from this persistent unit group.
   * @param unit The unit to remove from the group.
   * @return True if the unit was present in the group.
   */
  bool remove(BWAPI::Unit* unit)
  {
    const_iterator it = m_group.find(UnitWrapper(unit));
    const bool exists = (it != m_group.end());
    if (exists)
    {
      remove(it);
    }
    return exists;
  }

  /**
   * @brief Removes a unit from this persistent unit group.
   * @param pos The position of the unit to remove.
   */
  void remove(const_iterator pos)
  {
    BWAPI::Unit* unit = (*pos).getUnit();
    Spar->getUnitManager().unsubscribeIndividualUnitDestroy(unit, (*pos).m_unitDestroyID);
    m_group.erase(pos);
    m_unitRemovedEvent.raise(unit);

    if(m_group.empty())
    {
      m_emptyEvent.raise();
    }
  }

  /**
   * @brief Remove all the units from this group.
   */
  void clear()
  {
    iterator it = begin();
    iterator nextIt;
    while(it != end())
    {
      nextIt = it;
      ++nextIt;
      remove(it);
      it = nextIt;
    }
  }



  /**
   * @brief Checks whether the given unit is present in the group.
   * @param unit The unit to check for membership.
   * @return True if the unit is contained in the group.
   */
  bool contains(BWAPI::Unit* unit) const
  {
    return m_group.find(unit) != m_group.end();
  }

  const_iterator find(BWAPI::Unit* unit) const
  {
    return m_group.find(unit);
  }

  /**
   * @return The beginning of the sequence of units.
   */
  const_iterator begin() const
  {
    //return boost::transform_iterator<extract_unit, UnitWrapperSet::const_iterator>(m_group.begin());
    return m_group.begin();
  }

  /**
   * @return Past the end of the sequence of units.
   */
  const_iterator end() const
  {
    //return boost::transform_iterator<extract_unit, UnitWrapperSet::const_iterator>(m_group.end());
    return m_group.end();
  }

  /**
   * @return The beginning of the sequence of units.
   */
  iterator begin()
  {
    //return boost::transform_iterator<extract_unit, UnitWrapperSet::iterator>(m_group.begin());
    return m_group.begin();
  }

  /**
   * @return Past the end of the sequence of units.
   */
  iterator end()
  {
    //return boost::transform_iterator<extract_unit, UnitWrapperSet::iterator>(m_group.end());
    return m_group.end();
  }

  /**
   * @return The number of units contained in the group.
   */
  size_t size() const
  {
	  return m_group.size();
  }

  /**
   * @return Whether the group is empty.
   */
  bool empty() const
  {
    return m_group.size() == 0;
  }

  /**
   * @brief Notification when a unit is destroyed. The unit is removed from the group.
   * @param data The data specified when subscribing to the unit's destruction, a set iterator in this context.
   */
  void onUnitDestroy(void* data)
  {
    UnitWrapperSet::iterator* id = reinterpret_cast<UnitWrapperSet::iterator*>(data);
    remove(*id);
  }

  const std::string& getName() const { return m_name; }
  virtual std::string toString() const { return m_name; }

  friend std::ostream& operator<<(std::ostream& out, const PersistentUnitGroup& group);

  /**
   * @brief The OnEmpty event is raised when the group becomes empty.
   */
  DECLARE_EVENT(OnEmpty);
  OnEmpty& getEmptyEvent() const { return m_emptyEvent; }

  /**
   * @brief The OnGroupDestroyed event is raised when the group is destroyed (when it is not
   *        referenced any more).
   */
  DECLARE_EVENT(OnGroupDestroyed);
  OnGroupDestroyed& getGroupDestroyedEvent() const { return m_groupDestroyedEvent; }

  /**
   * @brief The OnUnitAdded event is raised when a unit is added to the group.
   */
  DECLARE_EVENT1(OnUnitAdded, BWAPI::Unit*);
  OnUnitAdded& getUnitAddedEvent() const { return m_unitAddedEvent; }

  /**
   * @brief The OnUnitRemoved event is raised when a unit is removed from the group.
   *        Note that the unit removed may have been destroyed.
   */
  DECLARE_EVENT1(OnUnitRemoved, BWAPI::Unit*);
  OnUnitRemoved& getUnitRemovedEvent() const { return m_unitRemovedEvent; }

  /**
   * @brief The empty group, included here for convenance.
   */
  static const PersistentUnitGroup EmptyGroup;

protected:
  PersistentUnitGroup(const PersistentUnitGroup& other);
  PersistentUnitGroup& operator=(const PersistentUnitGroup& other) const;

  EVENT(OnEmpty) m_emptyEvent;

  EVENT(OnGroupDestroyed) m_groupDestroyedEvent;

  EVENT(OnUnitAdded) m_unitAddedEvent;

  EVENT(OnUnitRemoved) m_unitRemovedEvent;

  const std::string m_name;

  EVENT_HANDLER(PersistentUnitGroup, onUnitDestroy) m_unitDestroyEventHandler;

  UnitWrapperSet m_group;

private:
  PersistentUnitGroup();
};