#pragma once
#include "../Utils/Component.h"
#include "../Utils/SafeMultiset.h"
#include "../Utils/Task.h"
#include <cassert>
#include <string>
#include <vector>
#include "timer.h"

/**
* @brief The Scheduler is responsible for executing tasks.
*/
class Scheduler
{
public:
  class InternalSchedulerTask;
  /**
   * @brief Compares two tasks, ordering by execution time then by priority.
   */
  struct TaskComparator
  {
    bool operator() (const InternalSchedulerTask* task1, const InternalSchedulerTask* task2) const;
  };
  typedef SafeMultiset<InternalSchedulerTask*, TaskComparator> TaskList;

public:
  /**
   * @brief The subscribers associates an ID to each activated task.
   */
  struct TaskID
  {
    TaskID()
      : m_isActive(false)
    {
    }
    /**
     * @brief Position of the task in the tasks multiset.
     */
    TaskList::iterator m_posInMultiset;
    /**
     * @brief Whether the task is activated (m_posInMultiset is valid iff the task is active).
     */
    bool m_isActive;
  };

  /**
   * @brief Task to be planned by the scheduler. This type is accessible to the outside. 
   *        Clients are expected to supply their own internal task 'memory' but should not modify
   *        the internal scheduler task's members.
   */
  class InternalSchedulerTask : public RunnableTask
  {
  public:
    /**
     * @brief Internal scheduler task constructor.
     * @param owner The owner of the task.
     */
    InternalSchedulerTask(const Component& owner)
      : RunnableTask(owner)
    {
    }

#ifdef _DEBUG_COMPONENT
    ~InternalSchedulerTask()
    {
      m_owner.onDeleteTask(this);
    }
#endif
    /**
     * @return Whether the task is active (registered in the scheduler).
     */
    bool isActive() const
    {
      return m_taskID.m_isActive;
    }

  private:
    friend class Scheduler;
    //
    // The following members are not visible to the client; they are only stored as part of the task
    // to eliminate the need to dynamically create them (to reduce the 'new' operations).
    //

    /**
     * @brief The ID associated with this task (not visible to the clients).
     */
    TaskID m_taskID;
    /**
     * @brief The priority of the task.
     */
    unsigned int m_priority;
    /**
     * @brief For periodic task, the number of frame between two executions, or 0 if the task is not periodic.
     */
    size_t m_period;
    /**
     * @brief Frame to execute to task at.
     */
    size_t m_executionFrame;
  };

public:
  static const int MILLISECOND_PER_FRAME = 40;

  Scheduler()
    : m_tasks(TaskComparator())
  {
  }

	/**
	* @brief The controller's destructor releases allocated resources.
	*/
	//~Scheduler();

  void initialize() {};

	/**
	 * @brief Executes the tasks that should be executed at or before the frame number.
	 */
	void onFrame()
  {
    //SPAR_LOG(LogTypes::GENERAL_ERROR, "%d, %d, %d", BWAPI::Broodwar->getLatencyTime(), BWAPI::Broodwar->getLatencyFrames(), BWAPI::Broodwar->getLatencyTime() / BWAPI::Broodwar->getLatencyFrames());
    // We need a better function to evaluate the time available at each frame
    m_tasks.do_while(ExecuteTask(*this, 99999999/*(Scheduler::MILLISECOND_PER_FRAME * Spar->getTimer().timeFrequency()) / 1000*/)); 
  }

	/**
	 * @brief Subscribes a new task to the controller.
	 * @param task The new task to subscribe.
   * @param priority The priority of the new task.
   * @param period The period of the new task, 0 if not periodic.
   * @param waitingTimeBeforeExecution The number of frames to wait before executing the task.
	 */
	void activate(InternalSchedulerTask* task, unsigned int priority, size_t period, size_t waitingTimeBeforeExecution)
  {
#ifdef _DEBUG_COMPONENT
    task->m_owner.onActivateTask(task);
#endif
    task->m_priority = priority;
    task->m_period = period;
    task->m_executionFrame = BWAPI::Broodwar->getFrameCount() + waitingTimeBeforeExecution;
    TaskList::iterator it = m_tasks.insert(task);
    task->m_taskID.m_posInMultiset = it;
    task->m_taskID.m_isActive = true;
  }

  /**
   * @brief Unregisters the task from the scheduler.
   * @param task The task to remove from the scheduler.
   */
  void deactivate(InternalSchedulerTask* task)
  {
    if (task->m_taskID.m_isActive)
    {
#ifdef _DEBUG_COMPONENT
      task->m_owner.onDeactivateTask(task);
#endif
      task->m_taskID.m_isActive = false;
      m_tasks.erase(task->m_taskID.m_posInMultiset);
    }
  }

  static const unsigned int SCHEDULING_MEDIUM_PRIORITY = 1000;
  static const unsigned int FSM_SCHEDULING_PRIORITY = SCHEDULING_MEDIUM_PRIORITY;

#ifdef _DEBUG_COMPONENT
  template <class Function>
  Function for_each_task(Function function) const
  {
    return m_tasks.do_while(function);
  }
#endif

private:
  /**
   * @brief Set of tasks to be executed.
	 */
  TaskList m_tasks;

  /**
   * @brief Functor to execute tasks.
   */
  class ExecuteTask
  {
  public:
    ExecuteTask(Scheduler& scheduler, Timer::value_type timeLimit)
      : m_scheduler(scheduler)
      , m_timeLimit(timeLimit)
    {
    }

    /**
     * @brief Executes the current task, copying it further in the list of the task
     *        is periodic.
     * @param taskWrapper The current task, including additional information like whether the task has been deleted.
     * @return Whether to break to iteration and whether to delete the current task.
     */
    TaskList::FunctionReturn operator()(TaskList::Wrapper& wrapper) const
      {
        //if(Spar->getTimer().currentTime() < m_timeLimit)
        //  return TaskList::FunctionReturn(false,true);

        InternalSchedulerTask* task = wrapper;

        if(static_cast<int>(task->m_executionFrame) > BWAPI::Broodwar->getFrameCount())
          return TaskList::FunctionReturn(false,true);
        
        const size_t period = task->m_period;

        // When running we want to allow the user to reactivate a non-periodic task
        if (period == 0)
        {
          m_scheduler.deactivate(task);
        }

        task->run();

        // At this point we have to know whether the user has deactivated explicitely the task (if it is periodic)
        
        if (period > 0)
        {
          // We do not use "task->m_taskID.m_isActive" since the user may have deactivated and restarted the task
          if (!wrapper.isDeleted())
          {
            m_scheduler.deactivate(task);

            m_scheduler.activate(task, task->m_priority, task->m_period, task->m_period);
          }
        }
        return TaskList::FunctionReturn(false,false);
      }
  protected:
    Scheduler& m_scheduler;
    Timer::value_type m_timeLimit;
  };
  friend class ExecuteTask;
};

inline bool Scheduler::TaskComparator::operator() (const InternalSchedulerTask* task1, const InternalSchedulerTask* task2) const
{
  return task1->m_executionFrame < task2->m_executionFrame
    || (task1->m_executionFrame == task2->m_executionFrame
      && task1->m_priority < task2->m_priority);
}