#pragma once
#include "Scheduler.h"
#include "../SparAIContainer.h"

/**
 * @brief Wrapper for scheduler tasks allowing parameter-less activation and deactivation.
 */
class SchedulerTask : protected Scheduler::InternalSchedulerTask
{
public:
/**
 * @brief Scheduler task constructor.
 * @param owner Owner of the task.
 * @param priority Priority of the task.
 * @param period Period of the task, 0 if not periodic.
 * @param waitingTimeBeforeExecution The number of frames to wait before the first execution of the task.
 */
  SchedulerTask(const Component& owner, unsigned int priority, size_t period, size_t waitingTimeBeforeExecution)
    : Scheduler::InternalSchedulerTask(owner)
    , m_priority(priority)
    , m_period(period)
    , m_waitingTimeBeforeExecution(waitingTimeBeforeExecution)
  {}

  /**
   * @returns Whether the task is scheduled for execution.
   */
  bool isActive() const
  {
    return Scheduler::InternalSchedulerTask::isActive();
  }

  /**
   * @brief Sets the number of frames to wait before execution.
   * @param waitingTimeBeforeExecution the number of frames to wait before execution.
   */
  void setWaitingTimeBeforeExecution(size_t waitingTimeBeforeExecution)
  {
    m_waitingTimeBeforeExecution = waitingTimeBeforeExecution;
  }

  /**
   * @brief Registers the task to the scheduler.
   */
  void activate()
  {
    assert(!isActive());
    Spar->getScheduler().activate(this, m_priority, m_period, m_waitingTimeBeforeExecution);
  }

  /**
   * @brief Unregisters the task from the scheduler.
   */
  void deactivate()
  {
    Spar->getScheduler().deactivate(this);
  }

protected:
  /**
   * @brief The priority of the task.
   */
  unsigned int m_priority;

  /** 
   * @brief Period of the task, 0 if not periodic.
   */
  size_t m_period;
  
  /**
   * @brief The number of frames to wait before the first execution of the task.
   */
  size_t m_waitingTimeBeforeExecution;

};

#define SCHEDULER_EVENT(Type, Function) \
  class SchedulerTask_##Type_##Function : public SchedulerTask \
  { \
  public: \
    SchedulerTask_##Type_##Function(Type& object, unsigned int priority, size_t period, size_t waitingTimeBeforeExecution) \
      : SchedulerTask(object, priority, period, waitingTimeBeforeExecution) \
      , m_object(object) {} \
    virtual ~SchedulerTask_##Type_##Function() {} \
  protected: \
    virtual void runV() \
    { \
      m_object.Function(); \
    } \
    Type& m_object; \
  }
