#pragma once
#include "SafeList.h"
#include "EventHandler.h"

/**
 * @brief Base class for events which other components may subscribe to or unsubscribe from.
 *        The virtual destructor is necessary only so that this class is virtual.
 */
struct EventBase
{
  virtual ~EventBase() {}

#ifdef _DEBUG_COMPONENT
  virtual void onOwnerDeleted(Component* owner) = 0;
#endif
};

/**
 * @brief Base class for events, templated on the type of arguments sent with the event.
 * @tparam Arguments Type of arguments, must be boost::tuple<...>
 */
template <class Arguments = boost::tuple<>>
class Event : public EventBase
{
public:
  /**
   * @brief The type of event handler that must be used by subscribers.
   */
  typedef EventHandlerT<Arguments> EventHandlerType;
  /**
   * @brief The type of ID given to the subscribers.
   */
  class SubscriberID
  {
  public:
    SubscriberID()
      : m_isValid(false)
    {}
    SubscriberID(typename SafeList<std::pair<EventHandlerType*, void*>>::iterator id)
      : m_id(id)
      , m_isValid(true)
    {}
    bool isValid() const
    {
      return m_isValid;
    }
  protected:
    friend class Event;
    /**
     * @brief Whether this subscription ID has been initialized.
     */
    bool m_isValid;
    /**
     * @brief The position of the subscriber in the list of subscribers.
     */
    typename SafeList<std::pair<EventHandlerType*, void*>>::iterator m_id;
  };

#ifndef _DEBUG_COMPONENT
  virtual ~Event() {}
#else
  virtual void onOwnerDeleted(Component* owner)
  {
    struct DeleteIfInvalidOwner
    {
      DeleteIfInvalidOwner(const Component* owner)
        : m_owner(owner)
      {
      }
      SafeListBase::FunctionReturn operator()(const std::pair<EventHandlerType*, void*>& p) const
      {
        bool shouldDelete = (&p.first->m_owner == m_owner);
        return SafeListBase::FunctionReturn(shouldDelete, false);
      }
    private:
      const Component* const m_owner;
    };

    m_subscribers.for_each(DeleteIfInvalidOwner(owner));
  }
 
  virtual ~Event()
  {
    struct Notify
    {
      Notify(Event* ev)
        : m_event(ev)
      {}
      SafeListBase::ConstFunctionReturn operator()(const std::pair<EventHandlerType*, void*>& p) const
      {
        p.first->m_owner.onEventDeleted(p.first, m_event);
        return SafeListBase::ConstFunctionReturn();
      }
    private:
      Event* const m_event;
    };

    m_subscribers.for_each(Notify(this));
  }
#endif

  /**
   * @brief Subscribes an event handler to this event.
   * @param eventHandler The handler to notify when this event is raised.
   * @param data Additional data associated with the subscriber which will be sent back upon raising the event.
   * @return The ID associated with the new subscription.
   */
  SubscriberID subscribe(EventHandlerType* eventHandler, void* data)
  {
#ifdef _DEBUG_COMPONENT
    eventHandler->m_owner.onEventSubscribe(eventHandler, this);
#endif
    return SubscriberID(m_subscribers.add(std::make_pair(eventHandler, data)));
  }

  /**
   * @brief Unsubscribes an event handler from this event.
   * @param id The ID associated with the subscription (the one given upon subscription).
   */
  void unsubscribe(SubscriberID& id)
  {
    if (id.m_isValid)
    {
#ifdef _DEBUG_COMPONENT
      std::pair<EventHandlerType*, void*>& pair = *id.m_id;
      pair.first->m_owner.onEventUnsubscribe(pair.first, this);
#endif
      m_subscribers.erase(id.m_id);
      id.m_isValid = false;
    }
  }

  /**
   * @brief Raises the event, which notifies the list of subscribers.
   * @param arguments The arguments associated with the event.
   */
  void internalRaise(Arguments& arguments)
  {
    struct RaiseOnEvent
    {
      RaiseOnEvent(Arguments& args) : m_args(args) { }
      typename SafeList<std::pair<EventHandlerType*, void*>>::FunctionReturn 
        operator()(std::pair<EventHandlerType*, void*>& pair)
      {
        pair.first->onEvent(pair.second, m_args);
        return typename SafeList<std::pair<EventHandlerType*, void*>>::FunctionReturn();
      }
    private:
      Arguments& m_args;
    };

    m_subscribers.for_each(RaiseOnEvent(arguments));
  }

protected:
  /**
   * @brief The list of subscribers to be notified when this event is raised.
   */
  SafeList<std::pair<EventHandlerType*, void*>> m_subscribers;
};

class Event0 : public Event<boost::tuple<>>
{
public:
  void raise()
  {
    internalRaise(boost::make_tuple());
  }
};

template <class T1>
class Event1 : public Event<boost::tuple<T1>>
{
public:
  void raise(T1 t1)
  {
    internalRaise(boost::make_tuple(t1));
  }
};

template <class T1, class T2>
class Event2 : public Event<boost::tuple<T1,T2>>
{
public:
  void raise(T1 t1, T2 t2)
  {
    internalRaise(boost::make_tuple(t1,t2));
  }
};

template <class T1, class T2, class T3>
class Event3 : public Event<boost::tuple<T1,T2,T3>>
{
public:
  void raise(T1 t1, T2 t2, T3 t3)
  {
    internalRaise(boost::make_tuple(t1,t2,t3));
  }
};

template <class T1, class T2, class T3, class T4>
class Event4 : public Event<boost::tuple<T1,T2,T3,T4>>
{
public:
  void raise(T1 t1, T2 t2, T3 t3, T4 t4)
  {
    internalRaise(boost::make_tuple(t1,t2,t3,t4));
  }
};

template <class T1, class T2, class T3, class T4, class T5>
class Event5 : public Event<boost::tuple<T1,T2,T3,T4,T5>>
{
public:
  void raise(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5)
  {
    internalRaise(boost::make_tuple(t1,t2,t3,t4,t5));
  }
};

#define DECLARE_EVENT(Name) \
  class Name : public Event0 \
  { \
  public: \
    virtual ~Name() {} \
  };

#define DECLARE_EVENT1(Name, T1) \
  class Name : public Event1<T1> \
  { \
  public: \
    virtual ~Name() {} \
  };

#define DECLARE_EVENT2(Name, T1, T2) \
  class Name : public Event2<T1,T2> \
  { \
  public: \
    virtual ~Name() {} \
  };

#define DECLARE_EVENT3(Name, T1, T2, T3) \
  class Name : public Event3<T1,T2,T3> \
  { \
  public: \
    virtual ~Name() {} \
  };

#define DECLARE_EVENT4(Name, T1, T2, T3, T4) \
  class Name : public Event4<T1,T2,T3,T4> \
  { \
  public: \
    virtual ~Name() {} \
  };

#define DECLARE_EVENT5(Name, T1, T2, T3, T4, T5) \
  class Name : public Event5<T1,T2,T3,T4,T5> \
  { \
  public: \
    virtual ~Name() {} \
  };

#define EVENT(Name) \
  mutable Name