#pragma once
#include "../../Utils/Utils.h"
#include <utility>
#include <boost/tuple/tuple.hpp>

class Component;

/**
 * @brief Base class for event handlers, which are receptors that listen to and handle events.
 *        The virtual destructor is necessary only so that this class is virtual.
 */
struct EventHandlerBase
{
  /**
   * @brief Constructs a new event handler. Event handlers are usually not instantiated directly
   *        with the 'new' operator; they are most often members of other classes.
   * @param owner The owner of this event handler, only used for debugging purposes.
   */
  EventHandlerBase(const Component& owner)
#ifdef _DEBUG_COMPONENT
    : m_owner(owner)
#endif
  {
    unused(owner);
  }

  virtual ~EventHandlerBase() {}

#ifdef _DEBUG_COMPONENT
  /**
   * @brief The owner of this event handler.
   */
  const Component& m_owner;
#endif
};

/**
 * @brief Base class for event handlers, templated on the type of arguments sent with the particular events.
 * @tparam Arguments Type of arguments, must be boost::tuple<...>
 */
template <class Arguments>
struct EventHandlerT : public EventHandlerBase
{
  /**
   * @brief Constructs a new event handler.
   * @param owner The owner of this event handler, only used for debugging purposes.
   */
  EventHandlerT(const Component& owner)
    : EventHandlerBase(owner)
  {}

  typedef Arguments arguments;

  /**
   * @brief Handles the event.
   * @param data Arbitrary data passed to the event upon registration.
   * @param args Arguments sent with the event.
   */
  virtual void onEvent(void* data, arguments& args) = 0;
};

/**
 * @brief Shorthand for event handlers with no argument.
 */
struct EventHandler : EventHandlerT<boost::tuple<>>
{
  EventHandler(const Component& owner)
    : EventHandlerT(owner)
  {}
};

/**
 * @brief Shorthand for event handlers with one argument.
 * @tparam T1 Type of first argument.
 */
template<class T1>
struct EventHandler1 : EventHandlerT<boost::tuple<T1>>
{
  EventHandler1(const Component& owner)
    : EventHandlerT(owner)
  {}
};

/**
 * @brief Shorthand for event handlers with two arguments.
 * @tparam T1 Type of first argument.
 * @tparam T2 Type of second argument.
 */
template<class T1, class T2>
struct EventHandler2 : EventHandlerT<boost::tuple<T1,T2>>
{
  EventHandler2(const Component& owner)
    : EventHandlerT(owner)
  {}
};

/**
 * @brief Shorthand for event handlers with three arguments.
 * @tparam T1 Type of first argument.
 * @tparam T2 Type of second argument.
 * @tparam T3 Type of third argument.
 */
template<class T1, class T2, class T3>
struct EventHandler3 : EventHandlerT<boost::tuple<T1,T2,T3>>
{
  EventHandler3(const Component& owner)
    : EventHandlerT(owner)
  {}
};

/**
 * @brief Shorthand for event handlers with four arguments.
 * @tparam T1 Type of first argument.
 * @tparam T2 Type of second argument.
 * @tparam T3 Type of third argument.
 * @tparam T4 Type of fourth argument.
 */
template<class T1, class T2, class T3, class T4>
struct EventHandler4 : EventHandlerT<boost::tuple<T1,T2,T3,T4>>
{
  EventHandler4(const Component& owner)
    : EventHandlerT(owner)
  {}
};

/**
 * @brief Shorthand for event handlers with five arguments.
 * @tparam T1 Type of first argument.
 * @tparam T2 Type of second argument.
 * @tparam T3 Type of third argument.
 * @tparam T4 Type of fourth argument.
 * @tparam T5 Type of fifth argument.
 */
template<class T1, class T2, class T3, class T4, class T5>
struct EventHandler5 : EventHandlerT<boost::tuple<T1,T2,T3,T4,T5>>
{
  EventHandler5(const Component& owner)
    : EventHandlerT(owner)
  {}
};

//------------------------------------------------------------------------------

#define EVENT_HANDLER_NAME(Type, Function) \
  EventHandler_##Type##_##Function

#define EVENT_HANDLER(Type, Function) \
  struct EVENT_HANDLER_NAME(Type, Function) : EventHandler \
  { \
    EVENT_HANDLER_NAME(Type, Function)(Type& object) \
      : EventHandler(object) \
      , m_object(object) {} \
    virtual ~EVENT_HANDLER_NAME(Type, Function)() {} \
    virtual void onEvent(void* data, arguments&) \
    { \
      m_object.Function(data); \
    } \
  private: \
    Type& m_object; \
  }

#define EVENT_HANDLER1(Type, Function, T1) \
  struct EVENT_HANDLER_NAME(Type, Function) : EventHandler1<T1> \
  { \
    EVENT_HANDLER_NAME(Type, Function)(Type& object) \
      : EventHandler1(object) \
      , m_object(object) {} \
    virtual ~EVENT_HANDLER_NAME(Type, Function)() {} \
    virtual void onEvent(void* data, arguments& args) \
    { \
      m_object.Function(data, args.get<0>()); \
    } \
  private: \
    Type& m_object; \
  }

#define EVENT_HANDLER2(Type, Function, T1, T2) \
  struct EVENT_HANDLER_NAME(Type, Function) : EventHandler2<T1,T2> \
  { \
    EVENT_HANDLER_NAME(Type, Function)(Type& object) \
      : EventHandler2(object) \
      , m_object(object) {} \
    virtual ~EVENT_HANDLER_NAME(Type, Function)() {} \
    virtual void onEvent(void* data, arguments& args) \
    { \
      m_object.Function(data, args.get<0>(), args.get<1>()); \
    } \
  private: \
    Type& m_object; \
  }

#define EVENT_HANDLER3(Type, Function, T1, T2, T3) \
  struct EVENT_HANDLER_NAME(Type, Function) : EventHandler3<T1,T2,T3> \
  { \
    EVENT_HANDLER_NAME(Type, Function)(Type& object) \
      : EventHandler3(object) \
      , m_object(object) {} \
    virtual ~EVENT_HANDLER_NAME(Type, Function)() {} \
    virtual void onEvent(void* data, arguments& args) \
    { \
      m_object.Function(data, args.get<0>(), args.get<1>(), args.get<2>()); \
    } \
  private: \
    Type& m_object; \
  }

#define EVENT_HANDLER4(Type, Function, T1, T2, T3, T4) \
  struct EVENT_HANDLER_NAME(Type, Function) : EventHandler4<T1,T2,T3,T4> \
  { \
    EVENT_HANDLER_NAME(Type, Function)(Type& object) \
      : EventHandler4(object) \
      , m_object(object) {} \
    virtual ~EVENT_HANDLER_NAME(Type, Function)() {} \
    virtual void onEvent(void* data, arguments& args) \
    { \
      m_object.Function(data, args.get<0>(), args.get<1>(), args.get<2>(), args.get<3>()); \
    } \
  private: \
    Type& m_object; \
  }

#define EVENT_HANDLER5(Type, Function, T1, T2, T3, T4, T5) \
  struct EVENT_HANDLER_NAME(Type, Function) : EventHandler5<T1,T2,T3,T4,T5> \
  { \
    EVENT_HANDLER_NAME(Type, Function)(Type& object) \
      : EventHandler5(object) \
      , m_object(object) {} \
    virtual ~EVENT_HANDLER_NAME(Type, Function)() {} \
    virtual void onEvent(void* data, arguments& args) \
    { \
      m_object.Function(data, args.get<0>(), args.get<1>(), args.get<2>(), args.get<3>(), args.get<4>()); \
    } \
  private: \
    Type& m_object; \
  }
