#pragma once
#include "IFSMState.h"
#include "Transition.h"
#include "FSM.h"
#include <boost/tuple/tuple.hpp>

/**
 * @brief A FSM state represents a state in a finite state machine. Each state stores code to execute
 *        when it becomes active, and transitions associated to events to get out of that state.
 * @tparam FSM The type of the enclosing FSM.
 * @tparam DerivedState The most derived base class of this FSM state. Used to allow static polymorphism.
 */
template <class FSM, class DerivedState>
class FSMState : public IFSMState
{
public:
  FSMState(FSM &context)
    : m_context(context)
    , m_earliestNextTransitionFrame(0)
  {
  }

  virtual ~FSMState()
  {
  }

  std::string getName() const
  {
    // Could be done during compilation ;-)
    std::string fullName(typeid(DerivedState).name());
    size_t begin = fullName.find_last_of("::");
    if (begin == std::string::npos)
      begin = 0;
    else
      ++begin;
    return fullName.substr(begin);
  }

  /**
   * @brief This method is called when the FSM must be terminated due to an external event
   * (for example a unit FSM may be destroyed when its associated unit dies).
   * Note that this method should normally not be overriden in derived classes.
   */
  virtual void externalTerminate()
  {
    onExit();
  }

  /**
   * @brief Runs the (hopefully short) code associated with this state.
   */
  void onEnter()
  {
    m_context.setCurrentState(this);

    static_cast<DerivedState*>(this)->onEnterImpl();

    if (m_context.isExecuting())
    {
      enableTransitions();
    }
  }

  /**
   * @brief This method is called when a transition (or a termination) sends the finite state machine
   * out of that state.
   * It can be overriden in subclasses, but don't forget to call \ref FSMState::onExit() !
   */
  void onExit()
  {
    assert(m_context.m_currentState != NULL);

    static_cast<DerivedState*>(this)->onExitImpl();

    disableTransitions();

    m_context.onExitState();

#ifdef _DEBUG_COMPONENT
    verifyPendingSubscriptions("ERROR FSMState::onExit(): Make sure to unsubscribe EventHandler \"%s\" from event \"%s\" when the FSM switches states.");

    verifyPendingTasks("ERROR FSMState::onExit(): Make sure to deactivate task \"%s\" when the FSM switches states.");
#endif

    m_stateExitEvent.raise();
  }

  size_t getEarliestNextTransitionFrame() const
  {
    return m_earliestNextTransitionFrame;
  }

  void setEarliestNextTransitionFrame(size_t earliestAllowedFrameNumber)
  {
    m_earliestNextTransitionFrame = earliestAllowedFrameNumber;
  }

protected:
  /**
   * @brief EventsHelper is used solely to loop through the compile-time array of events (meta-programming).
   */
  template <size_t size>
  struct EventsHelper
  {
    static void activate(const FSMState& state)
    {
      boost::get<size-1>(static_cast<const DerivedState&>(state).m_events).activate();
      EventsHelper<size-1>::activate(state);
    }
    static void deactivate(const FSMState& state)
    {
      boost::get<size-1>(static_cast<const DerivedState&>(state).m_events).deactivate();
      EventsHelper<size-1>::deactivate(state);
    }
  };
  /**
   * @brief EventsHelper base case.
   */
  template <>
  struct EventsHelper<0>
  {
    static void activate(const FSMState&) {}
    static void deactivate(const FSMState&) {}
  };

  /**
   * @brief Enables all transitions attached to this state.
   */
  void enableTransitions()
  {
    EventsHelper<boost::tuples::length<DerivedState::Events>::value>::activate(*this);
  }

  /**
   * @brief Disables all transitions attached to this state.
   */
  void disableTransitions()
  {
    EventsHelper<boost::tuples::length<DerivedState::Events>::value>::deactivate(*this);
  }

  // Problem: "Events m_events" and "Transitions m_transitions" cannot be stored in here as DerivedType is
  // not yet defined in FSMState<DerivedState> (so they must be defined in derived classes instead).

  /**
   * @brief The enclosing FSM storing data shared by all states.
   */
  FSM &m_context;

  /**
   * @brief Indicates the earliest next frame where a transition is allowed to take place.
   *        This is used to prevent multiple transitions from occuring simultaneously (in the same frame); 
   *        this may happen if for example two Transitions are notified sequentially about the same event.
   */
  size_t m_earliestNextTransitionFrame;

private:
  FSMState(const FSMState& other);
};
