#pragma once
#include "SparQtGUI/mainwindow.h"
#include <SparObserver.h>
#include <SparAIModule.h>
#include <QtGui/QApplication>
#include <QtGui/QCloseEvent>
#include <QtCore/QThread>
#include <boost/thread/mutex.hpp>

template <class S1, class S2, class S3, class S4, class D1, class D2, class D3, class D4>
class SparQtGUI : public SparObserver
{
protected:
  typedef SparAIModule<S1,S2,S3,S4,D1,D2,D3,D4> Spar;

public:
  static const int DEFAULT_REFRESH_NB_FRAMES = 20;

  SparQtGUI(Spar* spar)
    : m_spar(spar), m_application(NULL), m_mainWindow(NULL), m_threadGUI(this), m_refreshNbFrames(DEFAULT_REFRESH_NB_FRAMES)
  {
    m_threadGUI.start();
  }

  ~SparQtGUI()
  {
  }

  virtual void close()
  {
    // This callback takes place in the main StarCraft thread.

    m_threadGUI.closeNow();
    m_threadGUI.wait();
  }

  virtual void onStart()
  {
    // This callback takes place in the main StarCraft thread.

    onFrame();
  }

  virtual void onFrame()
  {
    // This callback takes place in the main StarCraft thread.

    if (!m_threadGUI.isRunning())
      return;

    // TODO: event-driven instead of polling every x frames

    if (BWAPI::Broodwar->getFrameCount() % m_refreshNbFrames == 0)
    {
      // TODO: should use signal/slot instead of events

      // Data must be either copied or reference-counted since stuff may have been deleted once the
      // GUI thread gets a hold of it.

      QApplication::postEvent(m_mainWindow, new UpdateCurrentProcessesEvent(m_spar));

      QApplication::postEvent(m_mainWindow, new UpdatePerceptualStateEvent(m_spar->getPerceptualState()));

      QApplication::postEvent(m_mainWindow, new UpdateLayer3Event(m_spar->getD3()));
    }

    {
      //boost::mutex::scoped_lock lock(m_selectedProcessesUnitsMutex);

      m_selectedProcessesUnitsMutex.lock();

      for (std::set<BWAPI::Unit*>::const_iterator it = m_selectedProcessesUnits.begin(); it != m_selectedProcessesUnits.end(); ++it)
      {
        BWAPI::Unit* unit = *it;
        BWAPI::UnitType type = unit->getType();
        BWAPI::Position pos = unit->getPosition();
        BWAPI::Broodwar->drawBoxMap(pos.x() - type.dimensionLeft() - 2, pos.y() - type.dimensionUp() - 2, 
                                    pos.x() + type.dimensionLeft() + 2, pos.y() + type.dimensionUp() + 2,
                                    BWAPI::Colors::Red);
      }

      m_selectedProcessesUnitsMutex.unlock();
    }
  }

private:
  class QThreadGUI : public QThread
  {
  public:
    QThreadGUI(SparQtGUI* gui)
      : QThread(NULL)
      , m_gui(gui)
    {
    }

    void closeNow()
    {
      if (isRunning())
      {
        QApplication::postEvent(m_gui->m_application, new QCloseEvent());
      }
    }

    void run()
    {
      int argc = 0;
      {
        MainWindow::initialize();

        QApplication application(argc, NULL);
        MainWindow mainWindow(m_gui->m_selectedProcessesUnits, m_gui->m_selectedProcessesUnitsMutex);

        m_gui->m_application = &application;
        m_gui->m_mainWindow = &mainWindow;

        mainWindow.show();
        mainWindow.lower();
        application.exec();

        m_gui->m_application = NULL;
        m_gui->m_mainWindow = NULL;
      }
    }
  private:
    SparQtGUI* m_gui;
  };
  friend class QThreadGUI;

  Spar* m_spar;
  int m_refreshNbFrames;

  boost::mutex m_selectedProcessesUnitsMutex;
  std::set<BWAPI::Unit*> m_selectedProcessesUnits;

  // These pointers should actually be accessed via a mutex, but laziness has won the battle for now
  QApplication* m_application;
  MainWindow* m_mainWindow;

  QThreadGUI m_threadGUI;
};