#pragma once
#include "ui_mainwindow.h"
#include "model/Process.h"
#include "model/Location.h"
#include "model/Tech.h"
#include <QtCore/QEvent>
#include <SparAIModule.h>
#include <DecisionMaking/Layer3Tactical/Script/Layer3Script.h>
#include <boost/thread/mutex.hpp>
#include <sstream>

namespace Ui {
  class MainWindow;
}

// TODO: scheduler snapshot

class MainWindow : public QMainWindow
{
  Q_OBJECT

public:
  explicit MainWindow(std::set<BWAPI::Unit*>& units, boost::mutex& unitsMutex);
  ~MainWindow();

  static void initialize();

  virtual bool event(QEvent* ev);

  static int UpdateCurrentProcessesEventType;
  static int UpdatePerceptualStateEventType;
  static int UpdateLayer3EventType;

public slots:
  void chkMapAnalysisValueChanged(int value) { Spar->getConfiguration().setDisplayMap(bool(value)); }
  void chkEnemyInfluenceValueChanged(int value) { Spar->getConfiguration().setDisplayEnemyInfluenceMap(bool(value)); }
  void chkSelfInfluenceValueChanged(int value) { Spar->getConfiguration().setDisplaySelfInfluenceMap(bool(value)); }
  void chkVisibilityInfluenceValueChanged(int value) { Spar->getConfiguration().setDisplayVisibilityInfluenceMap(bool(value)); }
  void chkStructureInfluenceValueChanged(int value) { Spar->getConfiguration().setDisplayStructureInfluenceMap(bool(value)); }
  void chkOwnUnitsActionsValueChanged(int value) { Spar->getConfiguration().setDisplayOwnUnitsActions(bool(value)); }

private:
  void updateCurrentProcessesUnits();

  std::set<BWAPI::Unit*>& m_units;
  boost::mutex& m_unitsMutex;

  Ui::MainWindow *ui;

  static bool s_initialized;
};

class UpdateCurrentProcessesEvent : public QEvent
{
public:
  UpdateCurrentProcessesEvent(const Process* root)
    : QEvent((QEvent::Type)MainWindow::UpdateCurrentProcessesEventType)
  {
    m_root = ProcessWrapper::createWrapper(root);
  }
  ~UpdateCurrentProcessesEvent()
  {
    delete m_root;
  }
  ProcessWrapper* getRoot() { return m_root; }
private:
  ProcessWrapper* m_root;
};

class UpdatePerceptualStateEvent : public QEvent
{
public:
  UpdatePerceptualStateEvent(const PerceptualState& perceptualState)
    : QEvent((QEvent::Type)MainWindow::UpdatePerceptualStateEventType)
  {
    const Map* map = perceptualState.getMap();
    m_root = LocationWrapper::createWrapper(map);

    std::pair<Techs::tech_const_iterator, Techs::tech_const_iterator> techs = Techs::getAllTechs();
    for (; techs.first != techs.second; ++techs.first)
    {
      const Tech* tech = *techs.first;
      m_techs.push_back(TechWrapper(tech));
    }

    // TODO: add upper_bound iterator to location's interface
    std::pair<Map::enemy_unit_const_iterator, Map::enemy_unit_const_iterator> enemyUnits = map->getEnemyUnits();
    for (; enemyUnits.first != enemyUnits.second; ++enemyUnits.first)
    {
      const EnemyUnit* enemyUnit = *enemyUnits.first;
      std::pair<std::map<BWAPI::UnitType, size_t>::iterator, bool> result = 
        m_enemyArmyComposition.insert(std::make_pair(enemyUnit->getType(), 0));
      if (result.second)
      {
        result.first->second = map->getNbEnemyUnits(enemyUnit->getType());
      }
    }
  }
  ~UpdatePerceptualStateEvent()
  {
    delete m_root;
  }
  LocationWrapper* getRoot() { return m_root; }
  const std::vector<TechWrapper>& getTechs() const { return m_techs; }
  const std::map<BWAPI::UnitType, size_t>& getEnemyArmyComposition() const { return m_enemyArmyComposition; }
private:
  LocationWrapper* m_root;
  std::vector<TechWrapper> m_techs;
  std::map<BWAPI::UnitType, size_t> m_enemyArmyComposition;
};

class UpdateLayer3Event : public QEvent
{
public:
  template <class Layer3>
  UpdateLayer3Event(const Layer3& layer3)
    : QEvent((QEvent::Type)MainWindow::UpdateLayer3EventType)
  {
    initialize(layer3);
  }
  const std::vector<QString>& getTargetArmyComposition() const { return m_targetArmyComposition; }
  const std::vector<QString>& getTargetTechs() const { return m_targetTechs; }
  const std::vector<QString>& getCurrentTechActions() const { return m_currentTechActions; }
private:
  template <class Layer3>
  void initializeBase(const Layer3& layer3)
  {
    const ArmyComposition& army = layer3.getTargetArmyComposition();
    const std::map<BWAPI::UnitType, size_t>& unlimitedComposition = army.getUnlimitedComposition();
    for (std::map<BWAPI::UnitType, size_t>::const_iterator it = unlimitedComposition.begin(); it != unlimitedComposition.end(); ++it)
    {
      const std::pair<BWAPI::UnitType, size_t>& unit = *it;
      std::stringstream ss;
      ss << unit.second << " " << unit.first;
      m_targetArmyComposition.push_back(QString::fromStdString(ss.str()));
    }

    const TechSet& techs = layer3.getTargetTechs();
    for (TechSet::const_iterator it = techs.begin(); it != techs.end(); ++it)
    {
      m_targetTechs.push_back(QString::fromStdString(getObjectDescription(*it)));
    }
  }

  template <class Layer3>
  void initialize(const Layer3& layer3)
  {
    initializeBase(layer3);
  }
  template <>
  void initialize<Layer3Script>(const Layer3Script& layer3)
  {
    initializeBase(layer3);
    const ProgressTechTask* progressTech = layer3.getProgressTechTask();
    std::pair<ProgressTechTask::TechActionIterator, ProgressTechTask::TechActionIterator> techActions = progressTech->getCurrentTechActions();
    for (; techActions.first != techActions.second; ++techActions.first)
    {
      std::pair<const Tech*, const Layer3Script::TacticalAction*> tech = *techActions.first;
      std::stringstream ss;
      ss << *tech.first << ": " << *tech.second;
      m_currentTechActions.push_back(QString::fromStdString(ss.str()));
    }
  }

  std::vector<QString> m_targetArmyComposition;
  std::vector<QString> m_targetTechs;
  std::vector<QString> m_currentTechActions;
};
