#pragma once
#include "../QTreeWidgetItemCustom.h"
#include <QtCore/QString>
#include <PerceptualState/PerceptualState.h>

class LocationWrapper
{
public:
  LocationWrapper(const Location* location)
    : m_location(location)
    , m_name(QString::fromStdString(getObjectDescription(location)))
  {
  }

  virtual ~LocationWrapper() {}

  void displayIn(QTreeWidgetItemKey* item) const
  {
    item->setKey(m_location);
    item->setText(0, m_name);

    displayInImpl(item);

    item->setExpanded(true);
  };

  static LocationWrapper* createWrapper(const Location* location);

  virtual size_t getNbUnits() const = 0;

protected:
  virtual void displayInImpl(QTreeWidgetItemKey* item) const
  {
    unused(item);
  }

  // Only the address is used, cause the pointer itself may have been deleted
  const Location* const m_location;

  const QString m_name;
};

class PrimitiveLocationWrapper : public LocationWrapper
{
public:
  class UnitWrapper
  {
  public:
    template <class UnitType>
    UnitWrapper(const UnitType* unit)
      : m_unit(unit)
      , m_name(QString::fromStdString(getObjectDescription(unit)))
      , m_accessible(unit->isAccessible())
      , m_position(unit->getPosition())
      , m_unitType(unit->getType())
    {}
    void displayIn(QTreeWidgetItemKey* item) const
    {
      item->setKey(m_unit);
      item->setText(0, m_name);

      displayInImpl(item);

      item->setExpanded(true);
    };
  protected:
    virtual void displayInImpl(QTreeWidgetItemKey* item) const
    {
      if (!m_accessible)
      {
        QFont font(item->font(0));
        font.setItalic(true);
        item->setFont(0, font);
      }
    }
    const UnitLocationData* const m_unit;
    const QString m_name;
    const bool m_accessible;
    const BWAPI::Position m_position;
    const BWAPI::UnitType m_unitType;
  };

  class OwnUnitWrapper : public UnitWrapper
  {
  public:
    OwnUnitWrapper(const OwnUnit* unit)
      : UnitWrapper(unit)
    {}
  protected:
    virtual void displayInImpl(QTreeWidgetItemKey* item) const
    {
      UnitWrapper::displayInImpl(item);
      QFont font(item->font(0));
      font.setBold(true);
      item->setFont(0, font);
    }
  };

  class EnemyUnitWrapper : public UnitWrapper
  {
  public:
    EnemyUnitWrapper(const EnemyUnit* unit)
      : UnitWrapper(unit)
    {}
  protected:
    virtual void displayInImpl(QTreeWidgetItemKey* item) const
    {
      UnitWrapper::displayInImpl(item);
      QFont font(item->font(0));
      font.setBold(true);
      item->setFont(0, font);
    }
  };

  class NeutralUnitWrapper : public UnitWrapper
  {
  public:
    NeutralUnitWrapper(const NeutralUnit* unit)
      : UnitWrapper(unit)
    {}
  };

  PrimitiveLocationWrapper(const PrimitiveLocation* location)
    : LocationWrapper(location)
  {
    std::pair<PrimitiveLocation::own_unit_const_iterator, PrimitiveLocation::own_unit_const_iterator> ownBuildings = location->getOwnUnits();
    for (; ownBuildings.first != ownBuildings.second; ++ownBuildings.first)
    {
      m_units.push_back(new OwnUnitWrapper(*ownBuildings.first));
    }

    std::pair<PrimitiveLocation::enemy_unit_const_iterator, PrimitiveLocation::enemy_unit_const_iterator> enemyBuildings = location->getEnemyUnits();
    for (; enemyBuildings.first != enemyBuildings.second; ++enemyBuildings.first)
    {
      m_units.push_back(new EnemyUnitWrapper(*enemyBuildings.first));
    }

    std::pair<PrimitiveLocation::neutral_unit_const_iterator, PrimitiveLocation::neutral_unit_const_iterator> neutralUnits = location->getNeutralUnits();
    for (; neutralUnits.first != neutralUnits.second; ++neutralUnits.first)
    {
      m_units.push_back(new NeutralUnitWrapper(*neutralUnits.first));
    }
  }
  ~PrimitiveLocationWrapper()
  {
    for (std::list<UnitWrapper*>::iterator it = m_units.begin(); it != m_units.end(); ++it)
    {
      delete *it;
    }
  }
  virtual size_t getNbUnits() const
  {
    return m_units.size();
  }
protected:
  virtual void displayInImpl(QTreeWidgetItemKey* item) const
  {
    LocationWrapper::displayInImpl(item);
    for (std::list<UnitWrapper*>::const_iterator it = m_units.begin(); it != m_units.end(); ++it)
    {
      const UnitWrapper* unit = *it;
      QTreeWidgetItemKey* childUnit = new QTreeWidgetItemKey(item);
      unit->displayIn(childUnit);
    }    
  }
  std::list<UnitWrapper*> m_units;
};

class CompositeLocationWrapper : public LocationWrapper
{
public:
  CompositeLocationWrapper(const CompositeLocation* location)
    : LocationWrapper(location)
  {
    std::pair<CompositeLocation::sub_location_const_iterator, CompositeLocation::sub_location_const_iterator> subLocations = location->getSubLocations();
    for (; subLocations.first != subLocations.second; ++subLocations.first)
    {
      const Location* childLocation = *subLocations.first;
      m_childLocations.push_back(LocationWrapper::createWrapper(childLocation)); 
    }
    struct GreaterNbUnits
    {
      bool operator()(const LocationWrapper* first, const LocationWrapper* second) const
      {
        return first->getNbUnits() > second->getNbUnits();
      }
    };
    std::sort(m_childLocations.begin(), m_childLocations.end(), GreaterNbUnits());
  }
  ~CompositeLocationWrapper()
  {
    for (std::vector<LocationWrapper*>::const_iterator it = m_childLocations.begin(); it != m_childLocations.end(); ++it)
    {
      delete *it;
    }
  }
  virtual size_t getNbUnits() const
  {
    size_t nb = 0;
    for (std::vector<LocationWrapper*>::const_iterator it = m_childLocations.begin(); it != m_childLocations.end(); ++it)
    {
      nb += (*it)->getNbUnits();
    }
    return nb;
  }
protected:
  virtual void displayInImpl(QTreeWidgetItemKey* item) const
  {
    LocationWrapper::displayInImpl(item);
    for (std::vector<LocationWrapper*>::const_iterator itLocation = m_childLocations.begin(); itLocation != m_childLocations.end(); ++itLocation)
    {
      const LocationWrapper* childLocation = *itLocation;
      QTreeWidgetItemKey* child = new QTreeWidgetItemKey(item);
      childLocation->displayIn(child);
    }
  }

  std::vector<LocationWrapper*> m_childLocations;
};
