#pragma once
#include "../QTreeWidgetItemCustom.h"
#include <QtCore/QString>
#include <Utils/Process.h>
#include <DecisionMaking/Layer1Reactive/FSM/Common.h>

class ProcessWrapper
{
public:
  ProcessWrapper(const Process* process)
    : m_process(process)
    , m_name(QString::fromStdString(getObjectDescription(process)))
  {
    std::pair<Process::ProcessIterator, Process::ProcessIterator> childProcesses = process->getCurrentChildProcesses();
    for (Process::ProcessIterator it = childProcesses.first; it != childProcesses.second; ++it)
    {
      const Process* childProcess = *it;
      m_childProcesses.push_back(ProcessWrapper::createWrapper(childProcess));
    }
  }

  virtual ~ProcessWrapper()
  {
    for (std::list<ProcessWrapper*>::const_iterator it = m_childProcesses.begin(); it != m_childProcesses.end(); ++it)
    {
      delete *it;
    }
  }

  void displayIn(QTreeWidgetItemCustom<const Process*>* item) const
  {
    item->setKey(m_process);
    item->setCustomData(m_process);
    item->setText(0, m_name);

    displayInImpl(item);

    item->setExpanded(true);
  };

  static ProcessWrapper* createWrapper(const Process* process);

protected:
  virtual void displayInImpl(QTreeWidgetItemCustom<const Process*>* item) const
  {
    for (std::list<ProcessWrapper*>::const_iterator it = m_childProcesses.begin(); it != m_childProcesses.end(); ++it)
    {
      const ProcessWrapper* childProcess = *it;
      QTreeWidgetItemCustom<const Process*>* child = new QTreeWidgetItemCustom<const Process*>(item);
      childProcess->displayIn(child);
    }
  }

  // Only the address is used, cause the pointer itself may have been deleted
  const Process* const m_process;

  const QString m_name;

  std::list<ProcessWrapper*> m_childProcesses;
};

class BehaviorWrapper : public ProcessWrapper
{
public:
  BehaviorWrapper(const IBehavior* behavior)
    : ProcessWrapper(behavior)
  {
    std::vector<const IFSMState*> states = behavior->getAllStates();
    const IFSMState* currentState = behavior->getCurrentState();
    for (std::vector<const IFSMState*>::const_iterator it = states.begin(); it != states.end(); ++it)
    {
      std::list<BehaviorStateWrapper>::iterator itState = m_states.insert(m_states.end(), BehaviorStateWrapper(*it, (*it) == currentState));
    }
  }

protected:
  virtual void displayInImpl(QTreeWidgetItemCustom<const Process*>* item) const
  {
    for (std::list<BehaviorWrapper::BehaviorStateWrapper>::const_iterator itState = m_states.begin(); itState != m_states.end(); ++itState)
    {
      const BehaviorWrapper::BehaviorStateWrapper& s = *itState;
      QTreeWidgetItemCustom<const Process*>* state = new QTreeWidgetItemCustom<const Process*>(item);
      state->setCustomData(m_process);
      if (s.isCurrentState())
        ProcessWrapper::displayInImpl(state);
      s.displayIn(state);
    }
  }
  class BehaviorStateWrapper
  {
  public:
    BehaviorStateWrapper(const IFSMState* state, bool isCurrentState)
      : m_name(QString::fromStdString(getObjectDescription(state)))
      , m_state(state)
      , m_isCurrentState(isCurrentState)
    {
    }
    void displayIn(QTreeWidgetItemKey* item) const
    {
      item->setKey(m_state);
      item->setText(0, m_name);
      if (m_isCurrentState)
      {
        QFont font(item->font(0));
        font.setBold(true);
        item->setFont(0, font);
      }
      item->setExpanded(true);
    }
    bool isCurrentState() const { return m_isCurrentState; }
  protected:
    const QString m_name;
    const IFSMState* const m_state;
    bool m_isCurrentState;
  };

  std::list<BehaviorStateWrapper> m_states;
};
